# -*- coding: utf-8 -*-

from typing import Any, Optional, TYPE_CHECKING

import logging

import warnings

import pathlib

from PySide6.QtCore import Qt, Slot, Signal

from PySide6.QtGui import QIcon, QPainter, QOpenGLContext, QOffscreenSurface, QMatrix4x4

from PySide6.QtOpenGL import (
    QOpenGLShaderProgram, QOpenGLShader, QOpenGLTexture, QAbstractOpenGLFunctions,
    QOpenGLVertexArrayObject
)

from shiboken6 import VoidPtr

from OpenGL import GL

from pyqtgraph.colormap import ColorMap

from OpenGL.constant import IntConstant

import numpy as np

import rasterio

from rasterio.vrt import WarpedVRT

from insarviz.map.TreeModel import TreeItem, TreeModel

from insarviz.map.TreeItemAttribute import (
    TreeItemAttribute, TreeItemColormapAttribute, TreeItemFloatAttribute
)

from insarviz.map.Shaders import (
    DATA_UNIT, COLORMAP_UNIT, VERT_SHADER, ALPHA_SHADER, COLORMAP_SHADER, IMAGE_RGB_SHADER
)

from insarviz.linalg import matrix

from insarviz.colormaps import my_colormaps, my_cyclic_colormaps, create_colormap_texture

from insarviz.Loader import Loader

from insarviz.ColormapWidget import ColormapWidget

if TYPE_CHECKING:
    from insarviz.map.MapModel import MapModel


logger = logging.getLogger(__name__)


class Layer(TreeItem):

    kind: str = "layer"  # description of the class, used in RemoveTreeItemCommand
    removable: bool = True
    renamable: bool = True
    icon: QIcon = QIcon()  # look https://github.com/qgis/QGIS/tree/master/images/themes/default

    request_paint = Signal()

    def __init__(self, name: str):
        super().__init__()
        # whether the layer is visible in MapView or not, also checkbox state in LayerView
        self.visible: bool = True
        self.name: str = name

    class ShowParams:
        def __init__(self):
            self.cutoff: float = 0.0
    def update_show_params(self, params) -> float:
        pass

    def show(self, view_matrix: matrix.Matrix, projection_matrix: matrix.Matrix, show_params: "Layer.ShowParams",
             painter: Optional[QPainter] = None, vao: Optional[QOpenGLVertexArrayObject] = None,
             glfunc: Optional[QAbstractOpenGLFunctions] = None, blend: bool = True) -> None:
        """
        Shall be implemented by subclasses.
        Display the layer using either OpenGL commands or painter.

        Parameters
        ----------
        view_matrix : matrix.Matrix
            Transform world coordinates into view coordinates.
        projection_matrix : matrix.Matrix
            Transform view coordinates into clip coordinates (OpenGL only)
        painter : QPainter, optional
            QPainter provided by the view. If painter is given then vao_id must be given aswell.
        vao_id : QOpenGLVertexArrayObject, optional
            OpenGL Vertex Array Object, the VAO is a square mapped with a texture.
        glfunc : QAbstractOpenGLFunctions, optional
            OpenGL functions
        blend : bool, optional
        """
        raise NotImplementedError

    def data(self, column: int, role: int) -> Any:
        if column == TreeModel.remove_column:
            return None
        if role in (Qt.ItemDataRole.DisplayRole, Qt.ItemDataRole.EditRole):
            return self.name
        if role == Qt.ItemDataRole.CheckStateRole:
            return Qt.CheckState.Checked if self.visible else Qt.CheckState.Unchecked
        if role == Qt.ItemDataRole.DecorationRole:
            return self.icon
        return None

    def flags(self, flags: Qt.ItemFlags, column: int) -> Qt.ItemFlags:
        if column == 0:
            # enable visibility checkbox and drag&drop
            flags = (flags | Qt.ItemFlag.ItemIsUserCheckable | Qt.ItemFlag.ItemIsDragEnabled |
                     Qt.ItemFlag.ItemIsDropEnabled)
            if self.renamable:
                # enable name edition
                flags = flags | Qt.ItemFlag.ItemIsEditable
        elif column == TreeModel.remove_column:
            if self.removable:
                # enable the remove button
                flags = flags | Qt.ItemFlag.ItemIsEditable
        return flags

    def set_data(self, value: Any, column: int, role: int) -> bool:
        if column != 0:
            return False
        if role == Qt.ItemDataRole.CheckStateRole:
            self.visible = (value == Qt.CheckState.Checked or value == Qt.CheckState.Checked.value)
            return True
        if role == Qt.ItemDataRole.EditRole and self.renamable:
            self.name = value
            return True
        return False

    def to_dict(self, project_path: pathlib.Path) -> dict[str, Any]:
        output: dict[str, Any] = {}
        output["kind"] = self.kind
        output["name"] = self.name
        output["visible"] = self.visible
        return output


class OpenGLLayer(Layer):

    # QOpenGLContext shared with every other contexts, set by ts_viz.MainWindow.initUI
    context: Optional[QOpenGLContext] = None
    # QOffscreenSurface to use OpenGL commands, set by ts_viz.MainWindow.initUI
    offscreen_surface: Optional[QOffscreenSurface] = None

    def __init__(self, name: str):
        super().__init__(name)
        # dict of texture_unit:texture
        self.textures: dict[int, QOpenGLTexture] = {}
        self.alpha: float = 1.
        self.add_child(TreeItemFloatAttribute(self, "alpha", name="alpha", vmin=0., vmax=1.,
                                              tooltip="transparency (between 0 and 1)"))
        assert self.context is not None, "OpengGLLayer.init: no OpenGL shared context"
        assert self.context.isValid(), "OpenGLLayer.init: OpenGL shared context not valid"
        assert self.offscreen_surface is not None, "OpengGLLayer.init: no QOffscreenSurface"
        assert self.offscreen_surface.isValid(), "OpenGLLayer.init: QOffscreenSurface not valid"
        self.context.makeCurrent(self.offscreen_surface)
        self.program = QOpenGLShaderProgram()
        self.build_program()
        self.context.doneCurrent()

    def build_program(self) -> None:
        """
        Shall be implemented by subclasses.
        Computes the OpenGL shader program used by the layer, possibly sets some constant uniforms
        to the program (such as texture unit for example) and then returns its id.
        A valid shared OpenGL context is supposed to already be current when this method is called.

        Note that texture unit shall be given as integer (without GL_TEXTURE0) with set_uniform.

        Returns
        -------
        program : int
            the OpenGL shader program id
        """
        # shall be implented by subclasses
        raise NotImplementedError

    @Slot(IntConstant, QOpenGLTexture)
    def set_texture(self, texture_unit: IntConstant, texture: QOpenGLTexture) -> None:
        """
        Set texture as value for key texture_unit in self.textures dictionnary.
        That information is used by self.show to bind the correct texture to the texture unit when
        displaying the layer.

        Parameters
        ----------
        texture_unit : int
            texture unit for glActiveTexture, shall be on the form GL_TEXTURE0 + ...
        texture : QOpenGLTexture
            texture to bind
        """
        self.textures[texture_unit] = texture

    def show(self, view_matrix: matrix.Matrix, projection_matrix: matrix.Matrix, show_params: Layer.ShowParams,
             painter: Optional[QPainter] = None, vao: Optional[QOpenGLVertexArrayObject] = None,
             glfunc: Optional[QAbstractOpenGLFunctions] = None, blend: bool = True) -> None:
        assert glfunc is not None
        if painter is not None:
            painter.beginNativePainting()
            # a VAO is required because QPainter bound its own VAO so we need to bind back our own
            assert vao is not None, "OpenGLLayer: vao is required when using QPainter"
        if blend:
            glfunc.glEnable(GL.GL_BLEND)
            glfunc.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA)
        if vao is not None:
            vao.bind()
        self.program.bind()
        # bind textures to texture units
        for texture_unit, texture in self.textures.items():
            glfunc.glActiveTexture(int(texture_unit))
            texture.bind()
        # set view and projection matrixes
        self.program.setUniformValue(self.program.uniformLocation('view_matrix'),
                                     QMatrix4x4(matrix.flatten(view_matrix)))
        self.program.setUniformValue(self.program.uniformLocation('projection_matrix'),
                                     QMatrix4x4(matrix.flatten(projection_matrix)))
        # set alpha value
        self.program.setUniformValue1f(self.program.uniformLocation('alpha'), self.alpha)
        # set cutoff value
        self.program.setUniformValue1f(self.program.uniformLocation('cutoff'), show_params.cutoff)
        # draw the two triangles of the VAO that form a square
        glfunc.glDrawElements(GL.GL_TRIANGLES, 6, GL.GL_UNSIGNED_INT, VoidPtr(0))
        for _, texture in self.textures.items():
            texture.release()
        self.program.release()
        if vao is not None:
            vao.release()
        if painter is not None:
            painter.endNativePainting()

    def __del__(self):
        """
        Free textures and shader program from the VRAM when the layer is destroyed to prevent
        memory leaks.
        """
        try:
            if self.context.isValid() and self.offscreen_surface.isValid():
                self.context.makeCurrent(self.offscreen_surface)
                # delete the OpenGL textures
                for _, texture in self.textures.items():
                    texture.destroy()
                # delete the OpenGL shaders program
                del self.program
                self.context.doneCurrent()
        except RuntimeError:
            # the context has already been deleted
            pass

    def to_dict(self, project_path: pathlib.Path) -> dict[str, Any]:
        output: dict[str, Any] = super().to_dict(project_path)
        output["alpha"] = self.alpha
        return output


class MainLayer(OpenGLLayer):

    removable: bool = False
    renamable: bool = False
    kind: str = "main layer"

    def __init__(self, map_model: "MapModel"):
        self.width: int = map_model.tex_width
        self.height: int = map_model.tex_height
        super().__init__("Main Layer")
        self.filepath = map_model.loader.filepath
        self.add_child(TreeItemAttribute(self, "filepath", tooltip=self.filepath, editable=False))
        map_model.texture_changed.connect(self.set_texture)
        map_model.v0_v1_changed.connect(self.set_v0_v1)

    def build_program(self) -> None:
        self.program.addShaderFromSourceCode(QOpenGLShader.ShaderTypeBit.Vertex, VERT_SHADER)
        self.program.addShaderFromSourceCode(QOpenGLShader.ShaderTypeBit.Fragment, ALPHA_SHADER)
        self.program.addShaderFromSourceCode(QOpenGLShader.ShaderTypeBit.Fragment, COLORMAP_SHADER)
        self.program.link()
        self.program.bind()
        model_matrix: matrix.Matrix = matrix.scale(self.width, self.height)
        self.program.setUniformValue(self.program.uniformLocation('model_matrix'),
                                     QMatrix4x4(matrix.flatten(model_matrix)))
        self.program.setUniformValue1i(self.program.uniformLocation('values'), DATA_UNIT)
        self.program.setUniformValue1i(self.program.uniformLocation('colormap'), COLORMAP_UNIT)
        self.program.setUniformValue1f(self.program.uniformLocation('v0'), 0.)
        self.program.setUniformValue1f(self.program.uniformLocation('v1'), 1.)
        self.program.release()

    def data(self, column: int, role: int) -> Any:
        if column == 0 and role == Qt.ItemDataRole.ToolTipRole:
            return "Main data layer, insar data cube"
        return super().data(column, role)

    # connected to MapModel.v0_v1_changed
    @Slot(float, float)
    def set_v0_v1(self, v0: float, v1: float) -> None:
        """Update v0 and v1."""
        assert self.context is not None
        assert self.context.isValid()
        assert self.offscreen_surface is not None
        assert self.offscreen_surface.isValid()
        self.context.makeCurrent(self.offscreen_surface)
        self.program.bind()
        self.program.setUniformValue1f(self.program.uniformLocation('v0'), float(v0))
        self.program.setUniformValue1f(self.program.uniformLocation('v1'), float(v1))
        self.program.release()
        self.context.doneCurrent()

    @classmethod
    def from_dict(cls, input_dict: dict[str, Any], map_model: "MapModel") -> "MainLayer":
        assert input_dict["kind"] == cls.kind
        layer = MainLayer(map_model)
        if "visible" in input_dict:
            layer.visible = input_dict["visible"]
        if "alpha" in input_dict:
            layer.alpha = input_dict["alpha"]
        return layer

    def __del__(self):
        """
        Free textures and shader program from the VRAM when the layer is destroyed to prevent
        memory leaks.
        """
        try:
            if self.context.isValid() and self.offscreen_surface.isValid():
                self.context.makeCurrent(self.offscreen_surface)
                # textures are deleted by MapModel
                # delete the OpenGL shaders program
                del self.program
                self.context.doneCurrent()
        except RuntimeError:
            # the context has already been deleted
            pass


class RasterLayer(OpenGLLayer):

    nb_band: int = 0

    def __init__(self, name: str, model: "MapModel", filepath: pathlib.Path, bands: list[int],
                 mask: Optional[int]):
        self.model_matrix: matrix.Matrix = matrix.scale(model.tex_width, model.tex_height)
        super().__init__(name)
        self.filepath = filepath
        with warnings.catch_warnings():
            # ignore RuntimeWarning for not georeferenced file
            warnings.filterwarnings("ignore", category=rasterio.errors.NotGeoreferencedWarning,
                                    message="Dataset has no geotransform, gcps, or rpcs. The identity matrix will be returned.")
            with rasterio.open(filepath) as file:
                assert model.loader.dataset is not None
                assert len(bands) == self.nb_band, (f"{self.__class__} requires {self.nb_band} "
                                                    f"bands but received {len(bands)}")
                assert ((file.crs is not None and model.loader.dataset.crs is not None) or
                        (file.shape == model.loader.dataset.shape)), (f"{self.__class__} requires a ",
                                                                      "geolocalized image or an image "
                                                                      "of same shape than the dataset")
                assert self.context is not None
                assert self.context.isValid()
                assert self.offscreen_surface is not None
                assert self.offscreen_surface.isValid()
                self.add_child(TreeItemAttribute(self, "filepath",
                               tooltip=str(filepath), editable=False))
                self._bands = bands
                self._mask = mask
                if mask is not None:
                    bands += [mask]
                # warp the image if geolocated
                if (file.crs is not None and model.loader.dataset.crs is not None):
                    with WarpedVRT(file, crs=model.loader.dataset.crs) as vrt:
                        img = np.empty((*vrt.shape, self.nb_band+1), dtype=np.float32)
                        data = vrt.read(bands, out_dtype=np.float32)
                        # transpose to change shape from (bands, rows, columns) to (rows, columns, bands)
                        img[:, :, :len(bands)] = np.transpose(data, [1, 2, 0])
                        # model matrix transforms coordinates inside the square (0,1) to coordinates in
                        # pixels relatively to the dataset
                        self.model_matrix = matrix.from_rasterio_Affine(
                            ~model.loader.dataset.transform
                            * vrt.transform
                            * rasterio.Affine.scale(*vrt.shape[::-1]))
                        self.context.makeCurrent(self.offscreen_surface)
                        self.program.bind()
                        self.program.setUniformValue(self.program.uniformLocation('model_matrix'),
                                                     QMatrix4x4(matrix.flatten(self.model_matrix)))
                        self.program.release()
                        self.context.doneCurrent()
                else:
                    img = np.empty((*file.shape, self.nb_band+1), dtype=np.float32)
                    data = file.read(bands, out_dtype=np.float32)
                    # transpose to change shape from (bands, rows, columns) to (rows, columns, bands)
                    img[:, :, :len(bands)] = np.transpose(data, [1, 2, 0])
                # if mask is missing
                if mask is None:
                    img[:, :, self.nb_band] = 1.
                # mask nodata values
                if all(x is not None for x in file.nodatavals):
                    img[:, :, self.nb_band][np.logical_or.reduce(
                        [img[:, :, i] == file.nodatavals[i] for i in range(self.nb_band)])] = 0.
                elif file.nodata is not None:
                    img[:, :, self.nb_band][np.logical_or.reduce(
                        [img[:, :, i] == file.nodata for i in range(self.nb_band)])] = 0.
                # transform from integer to float if needed
                for i, band in enumerate(bands):
                    if np.dtype(file.dtypes[file.indexes.index(band)]) == np.uint8:
                        img[:, :, i] = img[:, :, i] / 255
                # set the value of others band to 0 when mask is 0
                img[:, :, :self.nb_band][img[:, :, self.nb_band] == 0] = 0.
                # set the value of mask band to 1 when > 0
                img[:, :, self.nb_band][img[:, :, self.nb_band] > 0] = 1.
                # create the texture
                self.load_image(img)

    def load_image(self, img: np.ndarray):
        # shall be implented by subclasses
        raise NotImplementedError

    def to_dict(self, project_path: pathlib.Path) -> dict[str, Any]:
        output: dict[str, Any] = super().to_dict(project_path)
        try:
            output["filepath"] = self.filepath.relative_to(project_path.parent).as_posix()
        except ValueError:
            output["filepath"] = self.filepath.as_posix()
        output["bands"] = self._bands
        output["mask"] = self._mask
        return output


class Raster1BLayer(RasterLayer):

    icon: QIcon = QIcon()
    kind: str = "raster 1B layer"
    nb_band: int = 1
    outlier_threshold = Loader.outlier_threshold
    autorange_threshold = ColormapWidget.autorange_threshold

    def __init__(self, name: str, model: "MapModel", filepath: pathlib.Path, band: int = 1,
                 mask: Optional[int] = None):
        if self.icon.isNull():
            # cannot be initialized in the class declaration because:
            # "QIcon needs a QGuiApplication instance before the icon is created." see QIcon doc
            Raster1BLayer.icon = QIcon('icons:raster1B.svg')
        self.colormap: ColorMap = my_colormaps[[c.name for c in my_colormaps].index('greyscale')]
        self.colormap_v0: float = 0.  # minimum value for colorbar range (default 5 percentile)
        self.colormap_v1: float = 1.  # maximum value for colorbar range (default 95 percentile)
        super().__init__(name, model, filepath, [band], mask)
        with warnings.catch_warnings():
            # ignore RuntimeWarning for slices that contain only nans
            warnings.filterwarnings("ignore", category=rasterio.errors.NotGeoreferencedWarning,
                                    message="Dataset has no geotransform, gcps, or rpcs. The identity matrix will be returned.")
            with rasterio.open(filepath) as file:
                band_description = file.descriptions[file.indexes.index(band)]
                self.band: str = (f"{band} ({band_description})"
                                  if band_description is not None else
                                  f"{band}")
                self.mask: str
                if mask is None:
                    self.mask = "None"
                else:
                    mask_description = file.descriptions[file.indexes.index(mask)]
                    self.mask = (f"{mask} ({mask_description})"
                                 if mask_description is not None else
                                 f"{mask}")
        self.add_child(TreeItemAttribute(self, "band", tooltip=self.band, editable=False))
        self.add_child(TreeItemAttribute(self, "mask", tooltip=self.mask, editable=False))
        self.set_colormap(self.colormap)
        self.add_child(TreeItemColormapAttribute(self, "colormap", tooltip="colormap"))

    def load_image(self, img: np.ndarray):
        assert self.context is not None
        assert self.context.isValid()
        assert self.offscreen_surface is not None
        assert self.offscreen_surface.isValid()
        h, w, d = img.shape
        if d == 2:
            # FIX for issue #183 (texture bug on mac M1/M2/M3)
            data = np.ones((h, w, 3), dtype='float32')
            data[:, :, 0:2] = img
            img = data
        self.context.makeCurrent(self.offscreen_surface)
        texture = QOpenGLTexture(QOpenGLTexture.Target.Target2D)
        texture.setMagnificationFilter(QOpenGLTexture.Filter.Nearest)
        texture.setMinificationFilter(QOpenGLTexture.Filter.LinearMipMapLinear)
        texture.setWrapMode(QOpenGLTexture.WrapMode.ClampToEdge)
        texture.setSize(w, h)
        texture.setFormat(QOpenGLTexture.TextureFormat.RGB32F)
        texture.allocateStorage(QOpenGLTexture.PixelFormat.RGB, QOpenGLTexture.PixelType.Float32)
        texture.setData(QOpenGLTexture.PixelFormat.RGB, QOpenGLTexture.PixelType.Float32, img.data)
        texture.generateMipMaps()
        self.context.doneCurrent()
        assert texture.textureId() != 0, f"{self.name} layer: cannot load image texture in OpenGl"
        self.set_texture(GL.GL_TEXTURE0+DATA_UNIT, texture)
        # compute histogram
        self.histogram = self.compute_histogram(img)
        # compute v0 v1
        self.set_v0_v1(*self.autorange_from_hist(*self.histogram))

    def compute_histogram(self, img: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
        data = img[:, :, 0][img[:, :, 1] > 0]
        # build bins excluding the most extreme percentiles
        bins = np.histogram_bin_edges(data, bins="fd", range=np.percentile(data, [1, 99]))
        # add bins at the start and end for outliers
        bins = np.array([np.min(data), *bins, np.max(data)])
        return np.histogram(data, bins=bins)

    autorange_from_hist = ColormapWidget.autorange_from_hist

    def build_program(self) -> None:
        self.program.addShaderFromSourceCode(QOpenGLShader.ShaderTypeBit.Vertex, VERT_SHADER)
        self.program.addShaderFromSourceCode(QOpenGLShader.ShaderTypeBit.Fragment, ALPHA_SHADER)
        self.program.addShaderFromSourceCode(QOpenGLShader.ShaderTypeBit.Fragment, COLORMAP_SHADER)
        self.program.link()
        self.program.bind()
        self.program.setUniformValue(self.program.uniformLocation('model_matrix'),
                                     QMatrix4x4(matrix.flatten(self.model_matrix)))
        self.program.setUniformValue1i(self.program.uniformLocation('values'), DATA_UNIT)
        self.program.setUniformValue1i(self.program.uniformLocation('colormap'), COLORMAP_UNIT)
        self.program.setUniformValue1f(self.program.uniformLocation('v0'), float(self.colormap_v0))
        self.program.setUniformValue1f(self.program.uniformLocation('v1'), float(self.colormap_v1))
        self.program.release()

    @Slot(ColorMap)
    def set_colormap(self, colormap: ColorMap) -> None:
        assert self.context is not None
        assert self.context.isValid()
        assert self.offscreen_surface is not None
        assert self.offscreen_surface.isValid()
        self.colormap = colormap
        self.context.makeCurrent(self.offscreen_surface)
        old_texture = self.textures.get(GL.GL_TEXTURE0+COLORMAP_UNIT, None)
        if old_texture is not None:
            old_texture.destroy()
        colormap_texture = create_colormap_texture(colormap)
        self.context.doneCurrent()
        self.set_texture(GL.GL_TEXTURE0+COLORMAP_UNIT, colormap_texture)

    @Slot(float, float)
    def set_v0_v1(self, v0: float, v1: float) -> None:
        self.colormap_v0 = float(v0)
        self.colormap_v1 = float(v1)
        assert self.context is not None
        assert self.context.isValid()
        assert self.offscreen_surface is not None
        assert self.offscreen_surface.isValid()
        self.context.makeCurrent(self.offscreen_surface)
        self.program.bind()
        self.program.setUniformValue1f(self.program.uniformLocation('v0'), float(self.colormap_v0))
        self.program.setUniformValue1f(self.program.uniformLocation('v1'), float(self.colormap_v1))
        self.program.release()
        self.context.doneCurrent()

    def to_dict(self, project_path: pathlib.Path) -> dict[str, Any]:
        output: dict[str, Any] = super().to_dict(project_path)
        output["colormap"] = self.colormap.name
        output["colormap_v0"] = self.colormap_v0
        output["colormap_v1"] = self.colormap_v1
        return output

    @classmethod
    def from_dict(cls, input_dict: dict[str, Any], map_model: "MapModel") -> "Raster1BLayer":
        assert input_dict["kind"] == cls.kind
        assert "filepath" in input_dict
        name = input_dict.get("name", "raster1B layer")
        bands = input_dict.get("bands", [1])
        mask = input_dict.get("mask", None)
        layer = Raster1BLayer(name, map_model, input_dict["filepath"], bands[0], mask)
        if "visible" in input_dict:
            layer.visible = input_dict["visible"]
        if "alpha" in input_dict:
            layer.alpha = input_dict["alpha"]
        if "colormap" in input_dict:
            try:
                colormap = my_colormaps[[c.name for c in my_colormaps].index(
                    input_dict["colormap"])]
                layer.set_colormap(colormap)
            except ValueError:
                try:
                    colormap = my_cyclic_colormaps[[c.name for c in my_cyclic_colormaps].index(
                        input_dict["colormap"])]
                    layer.set_colormap(colormap)
                except ValueError:
                    pass
        if "colormap_v0" in input_dict and "colormap_v1" in input_dict:
            v0, v1 = input_dict["colormap_v0"], input_dict["colormap_v1"]
            if isinstance(v0, float) and isinstance(v1, float):
                layer.set_v0_v1(v0, v1)
            else:
                logger.warning(f"layer {name}: colormap_v0 and/or colormap_v1 are not float, thus"
                               " they are ignored")
        return layer


class RasterRGBLayer(RasterLayer):

    icon: QIcon = QIcon()
    kind: str = "raster RGB layer"
    nb_band: int = 3

    def __init__(self, name: str, model: "MapModel", filepath: pathlib.Path, R: int = 1, G: int = 2,
                 B: int = 3, mask: Optional[int] = None):
        if self.icon.isNull():
            # cannot be initialized in the class declaration because:
            # "QIcon needs a QGuiApplication instance before the icon is created." see QIcon doc
            RasterRGBLayer.icon = QIcon('icons:rasterRGB.svg')
        super().__init__(name, model, filepath, [R, G, B], mask)
        with warnings.catch_warnings():
            # ignore RuntimeWarning for slices that contain only nans
            warnings.filterwarnings("ignore", category=rasterio.errors.NotGeoreferencedWarning,
                                    message="Dataset has no geotransform, gcps, or rpcs. The identity matrix will be returned.")
            with rasterio.open(filepath) as file:
                R_description = file.descriptions[file.indexes.index(R)]
                self.R: str = (f"{R} ({R_description})"
                               if R_description is not None else
                               f"{R}")
                G_description = file.descriptions[file.indexes.index(G)]
                self.G: str = (f"{R} ({G_description})"
                               if G_description is not None else
                               f"{G}")
                B_description = file.descriptions[file.indexes.index(B)]
                self.B: str = (f"{B} ({B_description})"
                               if B_description is not None else
                               f"{B}")
                self.mask: str
                if mask is None:
                    self.mask = "None"
                else:
                    mask_description = file.descriptions[file.indexes.index(mask)]
                    self.mask = (f"{mask} ({mask_description})"
                                 if mask_description is not None else
                                 f"{mask}")
        self.add_child(TreeItemAttribute(self, "R", tooltip=self.R, editable=False))
        self.add_child(TreeItemAttribute(self, "G", tooltip=self.G, editable=False))
        self.add_child(TreeItemAttribute(self, "B", tooltip=self.B, editable=False))
        self.add_child(TreeItemAttribute(self, "mask", tooltip=self.mask, editable=False))

    def load_image(self, img: np.ndarray):
        assert self.context is not None
        assert self.context.isValid()
        assert self.offscreen_surface is not None
        assert self.offscreen_surface.isValid()
        h, w, d = img.shape
        self.context.makeCurrent(self.offscreen_surface)
        texture = QOpenGLTexture(QOpenGLTexture.Target.Target2D)
        texture.setMagnificationFilter(QOpenGLTexture.Filter.Linear)
        texture.setMinificationFilter(QOpenGLTexture.Filter.LinearMipMapLinear)
        texture.setWrapMode(QOpenGLTexture.WrapMode.ClampToEdge)
        texture.setSize(w, h)
        texture.setFormat(QOpenGLTexture.TextureFormat.RGBA32F)
        texture.allocateStorage(QOpenGLTexture.PixelFormat.RGBA, QOpenGLTexture.PixelType.Float32)
        texture.setData(QOpenGLTexture.PixelFormat.RGBA, QOpenGLTexture.PixelType.Float32, img.data)
        texture.generateMipMaps()
        self.context.doneCurrent()
        assert texture.textureId() != 0, f"{self.name} layer: cannot load image texture in OpenGl"
        self.set_texture(GL.GL_TEXTURE0, texture)

    def build_program(self) -> None:
        self.program.addShaderFromSourceCode(QOpenGLShader.ShaderTypeBit.Vertex, VERT_SHADER)
        self.program.addShaderFromSourceCode(QOpenGLShader.ShaderTypeBit.Fragment, ALPHA_SHADER)
        self.program.addShaderFromSourceCode(QOpenGLShader.ShaderTypeBit.Fragment, IMAGE_RGB_SHADER)
        self.program.link()
        self.program.bind()
        self.program.setUniformValue(self.program.uniformLocation('model_matrix'),
                                     QMatrix4x4(matrix.flatten(self.model_matrix)))
        self.program.setUniformValue1i(self.program.uniformLocation('image'), 0)
        self.program.release()

    @classmethod
    def from_dict(cls, input_dict: dict[str, Any], map_model: "MapModel") -> "RasterRGBLayer":
        assert input_dict["kind"] == cls.kind
        assert "filepath" in input_dict
        name = input_dict.get("name", "rasterRGB layer")
        bands = input_dict.get("bands", [1, 2, 3])
        mask = input_dict.get("mask", None)
        layer = RasterRGBLayer(name, map_model, input_dict["filepath"], bands[0], bands[1],
                               bands[2], mask)
        if "visible" in input_dict:
            layer.visible = input_dict["visible"]
        if "alpha" in input_dict:
            layer.alpha = input_dict["alpha"]
        return layer



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
""" MinimapView

This module handles the creation of the Minimap view (general image display
of a band data) and user interactions with it (pan and zoom).
Works with the module MapModel (as in a Model/View architecture).

Contains class:
* MinimapView
"""
# imports ###################################################################

from PySide6.QtCore import Qt, QSize, Signal, Slot

from PySide6.QtGui import (
    QMouseEvent, QWheelEvent, QMatrix4x4
)

from PySide6.QtOpenGL import QOpenGLShaderProgram, QOpenGLShader

from PySide6.QtWidgets import QWidget

from OpenGL import GL

from insarviz.map.AbstractMapView import AbstractMapView

from insarviz.map.MapModel import MapModel

from insarviz.map.layers.Layer import MainLayer

from insarviz.map.Shaders import VIEWPORTRECT_VERT_SHADER, VIEWPORTRECT_FRAG_SHADER

from insarviz.linalg import matrix

# MiniMapView ##################################################################


class MinimapView(AbstractMapView):
    """
    Minimap
    This is a general view of the data. A white rectangle
    shows the area cureently displayed in Map (zoom/pan synchronized).
    """

    CLICK = 1

    pan_map_view = Signal(float, float)
    zoom_map_view = Signal(float, float, float)
    set_center_map_view = Signal(float, float)

    def __init__(self, map_model: MapModel, parent: QWidget):
        super().__init__(map_model, parent)
        # model matrix of the rectangle figuring MapView's viewport (in data coordinate system)
        self.mapview_viewport_matrix = matrix.identity()
        self.program: QOpenGLShaderProgram
        self.setCursor(Qt.CursorShape.OpenHandCursor)

    def initializeGL(self) -> None:
        super().initializeGL()
        glfunc = self.context().functions()
        glfunc.glDisable(GL.GL_BLEND)
        # shaders for the rectangle figuring MapView's viewport
        self.program = QOpenGLShaderProgram()
        self.program.addShaderFromSourceCode(QOpenGLShader.ShaderTypeBit.Vertex,
                                             VIEWPORTRECT_VERT_SHADER)
        self.program.addShaderFromSourceCode(QOpenGLShader.ShaderTypeBit.Fragment,
                                             VIEWPORTRECT_FRAG_SHADER)
        self.program.link()
        self.program.bind()
        self.program.setUniformValue(self.program.uniformLocation('rect_color'), 1., 1., 1., 1.)
        self.program.release()

    def sizeHint(self) -> QSize:
        return QSize(100, 100)

    def paintGL(self) -> None:
        glfunc = self.context().functions()
        glfunc.glClear(GL.GL_COLOR_BUFFER_BIT)
        ####
        for layer in self.map_model.layer_model.layers:
            if isinstance(layer, MainLayer):
                layer.show(self.view_matrix, self.projection_matrix, MainLayer.ShowParams(), vao=self.vao, glfunc=glfunc,
                           blend=False)
        # white rectangle figuring MapView's viewport
        self.vao.bind()
        self.program.bind()
        self.program.setUniformValue(self.program.uniformLocation('model_matrix'),
                                     QMatrix4x4(matrix.flatten(self.mapview_viewport_matrix)))
        self.program.setUniformValue(self.program.uniformLocation('view_matrix'),
                                     QMatrix4x4(matrix.flatten(self.view_matrix)))
        self.program.setUniformValue(self.program.uniformLocation('projection_matrix'),
                                     QMatrix4x4(matrix.flatten(self.projection_matrix)))
        glfunc.glDrawArrays(GL.GL_LINE_LOOP, 0, 4)
        self.program.release()
        self.vao.release()

    # interaction

    def mousePressEvent(self, e: QMouseEvent) -> None:
        # check if data loaded
        if self.map_model.current_band_index is not None:
            if self.interaction == self.INTERACTIVE:
                # NOTE self.p0 does not match AbstractMapView.p0:
                # it is the last position of the mouse while pressed,
                # not the position of the last mouse press
                self.p0 = (*self.get_texture_coordinate(int(e.position().x()), int(e.position().y())),
                           int(e.position().x()), int(e.position().y()))
                self.p_prev = self.p0
                if e.button() == Qt.MouseButton.LeftButton:
                    self.interaction = self.CLICK
        super().mousePressEvent(e)

    def mouseMoveEvent(self, e: QMouseEvent) -> None:
        if self.interaction == self.CLICK or self.left_drag or self.right_drag:
            x_data0, y_data0, x_widget0, y_widget0 = self.p0
            x_data1, y_data1, x_widget1, y_widget1 = (*self.get_texture_coordinate(int(e.position().x()), int(e.position().y())),
                                                      int(e.position().x()), int(e.position().y()))
            dx, dy = x_data1-x_data0, y_data1-y_data0
            if self.interaction == self.CLICK:
                # if the cursor has moved away from the press position buffer zone
                if abs(x_widget1 - x_widget0) + abs(y_widget1 - y_widget0) > 4:
                    self.interaction = self.INTERACTIVE
                    self.setCursor(Qt.CursorShape.ClosedHandCursor)
            elif self.left_drag:
                self.pan_map_view.emit(dx, dy)
                # NOTE self.p0 does not match AbstractMapView.p0:
                # it is the last position of the mouse while pressed,
                # not the position of the last mouse press
                self.p0 = x_data1, y_data1, x_widget1, y_widget1
            elif self.right_drag:
                # make the difference in view coordinates between the cursor previous
                # and current positions and zoom accordingly
                _, _, x_prev, y_prev = self.p_prev
                self.zoom_map_view.emit((x_widget1 - x_prev) -
                                        (y_widget1 - y_prev), x_data0, y_data0)
                # update previous positions
                self.p_prev = x_data1, y_data1, x_widget1, y_widget1

    def mouseReleaseEvent(self, e: QMouseEvent) -> None:
        if self.interaction == self.CLICK and e.button() == Qt.MouseButton.LeftButton:
            x0, y0, _, _ = self.p0
            self.set_center_map_view.emit(x0, y0)
            self.interaction = self.INTERACTIVE
        self.setCursor(Qt.CursorShape.OpenHandCursor)
        super().mouseReleaseEvent(e)

    def wheelEvent(self, event: QWheelEvent) -> None:
        x, y = self.get_texture_coordinate(int(event.position().x()), int(event.position().y()))
        ds = event.angleDelta().y()/8  # degrees
        self.zoom_map_view.emit(ds, x, y)

    # connected to MapView's viewport_matrix_changed
    @Slot(object)
    def update_mapview_viewport_matrix(self, m: matrix.Matrix) -> None:
        self.mapview_viewport_matrix = m
        self.update()

    def update_view_matrix(self) -> None:
        if self.map_model.tex_width and self.map_model.tex_height:
            # texture size
            tex_w, tex_h = self.map_model.tex_width, self.map_model.tex_height
            self.cx, self.cy = tex_w//2, tex_h//2
            # view size
            view_w, view_h = self.width(), self.height()
            # size ratios
            view_ratio, tex_ratio = view_w / view_h, tex_w / tex_h
            if view_ratio > tex_ratio:
                # view is wider than texture
                # zoom is thus the height ratio between view and texture
                self.z = view_h / tex_h
            else:
                # view is higher than texture
                # zoom is thus the width ratio between view and texture
                self.z = view_w / tex_w
        super().update_view_matrix()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" myCSVExporter

This module contains a custom class myCSVExporter that manages the format
of data export from plots to CSV file.
Contains class:
    * myCSVExporter - export plot data to custom csv file
and method:
    * dt_to_dec - convert a datetime to decimal year
"""

import logging
import datetime
import csv
import itertools


import pyqtgraph as pg

import insarviz.plot.TemporalPlotView
import insarviz.plot.SpatialPlotView

logger = logging.getLogger(__name__)


class myCSVExporter(pg.exporters.CSVExporter):
    Name = "CSV"

    def parameters(self):
        return None

    def export(self, fileName=None):
        if not isinstance(self.item, pg.PlotItem):
            raise TypeError("Must have a PlotItem selected for CSV export.")
        if fileName is None:
            self.fileSaveDialog(filter="*.csv")
            return None
        try:
            self.item.insarviz_widget
        except AttributeError:
            return super().export(fileName)
        if isinstance(self.item.insarviz_widget, insarviz.plot.TemporalPlotView.TemporalPlotWidget):
            return self.temporal_plot_export(fileName)
        if isinstance(self.item.insarviz_widget, insarviz.plot.SpatialPlotView.SpatialPlotWidget):
            return self.spatial_plot_export(fileName)
        return super().export(fileName)

    def temporal_plot_export(self, fileName):
        header = []
        data_columns = []
        unit = str(self.item.insarviz_widget.plot_model.loader.units)
        header.append("Band")
        data_columns.append(self.item.insarviz_widget.plot_model.loader.dataset.indexes)
        if isinstance(self.item.getAxis("bottom"), pg.DateAxisItem):
            dates = self.item.insarviz_widget.plot_model.loader.dates
            header.append("Date (YYYY-MM-DD)")
            data_columns.append([d.strftime("%Y-%m-%d") for d in dates])
            header.append("Date (decimal year)")
            data_columns.append([dt_to_dec(d) for d in dates])
        for k, item in enumerate(self.item.items):
            if hasattr(item, 'implements') and item.implements('plotData'):
                if hasattr(item, 'getOriginalDataset'):
                    # try to access unmapped, unprocessed data
                    item_data = item.getOriginalDataset()
                else:
                    # fall back to earlier access method
                    item_data = item.getData()
                if item_data[0] is None:
                    # no data found, break out...
                    continue
                if len(item_data[0]) != len(data_columns[0]):
                    raise RuntimeError("curve has a number of point unequal to the number of bands")
                data_columns.append(item_data[1])
                if item.name() is not None:
                    header.append(str(item.name()).replace('"', '""') + " (" + unit + ")")
                else:
                    header.append(f"curve {k}")
        with open(fileName, 'w', newline='', encoding="utf-8") as csvfile:
            writer = csv.writer(csvfile, delimiter=",", quoting=csv.QUOTE_MINIMAL)
            writer.writerow(header)
            for row in itertools.zip_longest(*data_columns, fillvalue=""):
                writer.writerow(row)

    def spatial_plot_export(self, fileName):
        header = []
        data_columns = []
        profile_curve_found = False
        unit = str(self.item.insarviz_widget.plot_model.loader.units)
        for item in self.item.items:
            if hasattr(item, 'implements') and item.implements('plotData'):
                if not profile_curve_found:
                    profile_curve_found = True
                    if hasattr(item, 'getOriginalDataset'):
                        # try to access unmapped, unprocessed data
                        item_data = item.getOriginalDataset()
                    else:
                        # fall back to earlier access method
                        item_data = item.getData()
                    if item_data[0] is None:
                        # no data found, break out...
                        logger.warning("CSV exporter: Empty profile curve")
                        item_data = [[], []]
                    bottom_axis = self.item.getAxis("bottom")
                    header.append(
                        f"{bottom_axis.labelText} ({bottom_axis.labelUnits})")
                    data_columns.append(item_data[0])
                    header.append(str(item.name()).replace('"', '""') + " (" + unit + ")")
                    data_columns.append(item_data[1])
                else:
                    raise RuntimeError("two curves")
        with open(fileName, 'w', newline='', encoding="utf-8") as csvfile:
            writer = csv.writer(csvfile, delimiter=",", quoting=csv.QUOTE_MINIMAL)
            writer.writerow(header)
            for row in itertools.zip_longest(*data_columns, fillvalue=""):
                writer.writerow(row)


def dt_to_dec(dt):
    """Convert a datetime to decimal year. Time is at beginning of day.
    from https://stackoverflow.com/questions/29851357/python-datetime-to-decimal-year-one-day-off-where-is-the-bug"""
    year_start = datetime.datetime(dt.year, 1, 1)
    year_end = year_start.replace(year=dt.year+1)
    return dt.year + ((dt - year_start).total_seconds() /  # seconds so far
                      float((year_end - year_start).total_seconds()))  # seconds in year



# -*- coding: utf-8 -*-

"""
Inspired from:
    - https://gis.stackexchange.com/questions/392695/is-it-possible-to-build-a-vrt-file-from-multiple-files-with-rasterio
    - https://pypi.org/project/rio-vrt/
"""

from typing import Union, Optional

import logging

import xml.etree.ElementTree
from xml.etree.ElementTree import Element, ElementTree, SubElement, indent

import datetime

import rasterio
from rasterio.shutil import copy as riocopy
from rasterio.io import MemoryFile

logger = logging.getLogger(__name__)


def build_vrt(src_files: Union[str, list[str]], bands: Union[None, int, list[int], list[list[int]]],
              out_filename: str,  dates: Optional[list[Optional[int]]] = None,
              value_unit: Optional[str] = None, scale_factor: Optional[float] = None) -> bool:
    """
    Build a VRT (Virtual Raster Dataset) from one or several files and one or several band numbers.
    Dates can optionally be provided, one date per band to extract (in %Y%m%d format).
    A value unit string can also be optionnally provided.
    Finaly, a scale factor can be provided, to be applied to every extracted band.
    Return True if the VRT was created successfully, and False otherwise.

    Parameters
    ----------
    src_files : Union[str, list[str]]
        List of files from which to take bands to build the VRT. If a single file and a list of
        bands are given, then the VRT consists of all the given bands from the given file.
    bands : Union[None, int, list[int], list[list[int]]]
        A list of list of band numbers, one list of band numbers for each source file, to
        extract to build the VRT. If None is given, all bands of all source_files are extracted.
        If a single band number is given, that band will be extracted from each source file. If a
        single source file and a list of band numbers is given, then those bands will be extracted
        from the source file.
    out_filename : str
        A relative path toward the filename of the output VRT (will be created if needed).
    dates : Optional[list[Optional[int]]], optional
        A list of dates in format %Y%m%d, one date per band to extract (in the same order).
        For example, for 30th July 2018 : "20180730"
        see https://docs.python.org/3/library/datetime.html#strftime-and-strptime-format-codes
    value_unit : Optional[str], optional
        String to use as the unit name of the bands (for example "radian").
    scale_factor : Optional[float], optional
        A scale factor to be applied to every extracted band

    Returns
    -------
    bool
        True if the VRT has been created, False if there was an problem along the way.
    """
    if scale_factor is not None:
        scale_factor = float(scale_factor)
        assert isinstance(scale_factor, float), "scale_factor must be float"
    if isinstance(src_files, str):
        src_files = [src_files]
    # from now on src_files is a list[str]
    if len(src_files) == 0:
        logger.error("There should be at least 1 file to create a vrt.")
        return False
    if bands is None:
        bands = []
        for src in src_files:
            with rasterio.open(src) as dataset:
                bands.append(list(dataset.indexes))
    elif isinstance(bands, int):
        bands = [[bands] for _ in src_files]
    elif all(isinstance(band, int) for band in bands):
        if len(src_files) == 1:
            bands = [bands]
        else:
            bands = [[band] for band in bands]
    # from now on bands is a list[list[int]]
    # meaning one list of band numbers to extract for each source file of src_files
    if len(src_files) != len(bands):
        logger.error("number of scr_files and number of bands not matching, abort")
        return False
    if dates is None:
        dates = [None for band in bands for _ in band]
    if sum(len(_) for _ in bands) != len(dates):
        logger.error("number of bands and number of dates not matching, abort")
        return False
    # perform sanity check that all source files have the same width, height, crs and transform
    # using set to store the values of the files and check that the length of the set is 1
    # (meaning that a single value has been found across all files)
    width, height = set(), set()
    crs = set()
    transform = set()
    for src in src_files:
        with rasterio.open(src) as dataset:
            width.add(dataset.width)
            height.add(dataset.height)
            if dataset.crs is None:
                crs.add(None)
            else:
                crs.add(dataset.crs.to_string())
            transform.add(dataset.transform)
    if len(width) != 1:
        logger.error(f"conflicting dataset widths found {width}, abort")
        return False
    rasterXSize = width.pop()
    if len(height) != 1:
        logger.error(f"conflicting dataset heights found {height}, abort")
        return False
    rasterYSize = height.pop()
    if len(crs) != 1:
        logger.error(f"conflicting dataset crs found {crs}, abort")
        return False
    if len(transform) != 1:
        logger.error("conflicting dataset georeferencing transformation matrixes"
                     f" found {transform}, abort")
        return False
    # create the list of bands to build the vrt
    out_bands = []
    # iterate over src_files and bands, adding targeted bands to out_bands
    for src, target_bands in zip(src_files, bands):
        with rasterio.open(src) as dataset, MemoryFile() as memfile:
            # copy the dataset in the memoryfile using VRT driver
            riocopy(dataset, memfile.name, driver='VRT')
            # parse the content of the memoryfile into an xml tree
            vrt_xml: Element = xml.etree.ElementTree.fromstring(memfile.read().decode('utf-8'))
            for target_band in target_bands:
                # select all children elements that are VRTRasterBand and have an attribute band
                # equal to target_band
                found_bands = vrt_xml.findall(f"VRTRasterBand[@band='{target_band}']")
                if len(found_bands) == 0:
                    logger.error(f"did not find band {target_band} in file {src}, abort")
                    return False
                if len(found_bands) > 1:
                    logger.error(f"found multiple band {target_band} in file {src}, abort")
                    return False
                found_band = found_bands[0]
                out_bands.append(found_band)
    # create the root of the VRT XML
    out_xml = Element('VRTDataset', {"rasterXSize": str(rasterXSize),
                                     "rasterYSize": str(rasterYSize)})
    # if crs is not None, give the spatial reference system from the first source file to the output
    if crs.pop() is not None:
        with rasterio.open(src_files[0]) as dataset, MemoryFile() as memfile:
            # copy the dataset in the memoryfile using VRT driver
            riocopy(dataset, memfile.name, driver='VRT')
            # parse the content of the memoryfile into an xml tree
            vrt_xml = xml.etree.ElementTree.fromstring(memfile.read().decode('utf-8'))
            # add the SRS and GeoTransform subelements
            out_xml.append(vrt_xml.find("SRS"))
            out_xml.append(vrt_xml.find("GeoTransform"))
            logger.info(f"CRS and geotransorm from {src_files[0]} set as CRS and geotransform")
    # if value_unit is not None, add it to the output
    if value_unit is not None:
        if out_xml.find("Metadata") is None:
            SubElement(out_xml, "Metadata")
        if out_xml.find("Metadata").find("MDI[@key='VALUE_UNIT']") is None:
            SubElement(out_xml.find("Metadata"), "MDI", attrib={"key": "VALUE_UNIT"})
        out_xml.find("Metadata").find("MDI[@key='VALUE_UNIT']").text = str(value_unit)
        logger.info(f"{value_unit} set as unit name")
    # add each band to the VRT XML
    for band_number, (band, date) in enumerate(zip(out_bands, dates), start=1):
        # change the band number from the one of the src file to the one in the built vrt
        band.set("band", str(band_number))
        #  if date is not provided, check if the end of the description is a date
        if date is None and band.find("Description") is not None:
            try:
                description_date = band.find("Description").text[-8:]
                datetime.datetime.strptime(description_date, "%Y%m%d")
                date = description_date
            except ValueError:
                pass
        # if date is provided update/add the DATE tag
        if date is not None:
            if band.find("Metadata") is None:
                SubElement(band, "Metadata")
            if band.find("Metadata").find("MDI[@key='DATE']") is None:
                SubElement(band.find("Metadata"), "MDI", attrib={"key": "DATE"})
            band.find("Metadata").find("MDI[@key='DATE']").text = str(date)
        # if scale_factor then change the source from SimpleSource to ComplexSource with ScaleRatio
        if scale_factor is not None:
            complex_source = band.find("SimpleSource")
            complex_source.tag = "ComplexSource"
            if band.find("NoDataValue") is not None:
                SubElement(complex_source, "NODATA")
                complex_source.find("NODATA").text = band.find("NoDataValue").text
            SubElement(complex_source, "ScaleRatio")
            complex_source.find("ScaleRatio").text = str(scale_factor)
            SubElement(complex_source, "ScaleOffset")
            complex_source.find("ScaleOffset").text = str(0)
        out_xml.append(band)
    # indent the VRT XML
    indent(out_xml)
    # write the VRT XML in the out file
    ElementTree(out_xml).write(out_filename, encoding="utf-8")
    logger.info(f"VRT file {out_filename} of {sum(len(_) for _ in bands)} bands created.")
    return True



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Mar 26 15:52:25 2024

@author: thomcoli
"""

from enum import IntEnum, auto

from PySide6.QtCore import Qt


class Roles(IntEnum):
    # used by Selection.SelectionFolder and LayerView.ProxyLayerModel
    FolderRole = Qt.ItemDataRole.UserRole
    # used by Selection.SelectionItem
    ShowCurveRole = auto()
    # used by Selection.SelectionProfile
    ProfileTemporalRole = auto()
    ProfileSpatialRole = auto()
    # used by LayerModel.TreeItemAttribute
    EditorRole = auto()  # used by LayerView.ItemDelegate
    DataRole = auto()
    # used by LayerModel.TreeItemAttribute as DataRole and PlotModel/PlotView
    ComputeDataRole = auto()
    CurveColorRole = auto()
    # used for a persistent editor for slider attributes
    PersistentEditorWidgetRole = auto()
    # Get the source dataset of a layer
    DatasetRole = auto()



# -*- coding: utf-8 -*-

from typing import Optional

from PySide6.QtCore import Qt, Slot

from PySide6.QtWidgets import (
    QDialog, QWidget, QDialogButtonBox, QVBoxLayout, QLabel, QComboBox, QCheckBox, QGridLayout
)


class RasterLayerDialog(QDialog):

    def __init__(self, input_bands: list[str], nb_output_bands: int,
                 output_bands_names: list[str], parent: Optional[QWidget] = None):
        assert len(output_bands_names) == nb_output_bands
        super().__init__(parent)
        self.setWindowModality(Qt.WindowModality.ApplicationModal)
        self.setModal(True)
        self.setWindowTitle("Choose 1B layer bands")
        # band comboboxes and labels
        self.band_comboboxes = []
        self.band_labels = []
        for i in range(nb_output_bands):
            self.band_comboboxes.append(QComboBox())
            self.band_comboboxes[i].addItems(input_bands)
            if len(input_bands) in (nb_output_bands, nb_output_bands+1):
                self.band_comboboxes[i].setCurrentIndex(i)
            else:
                self.band_comboboxes[i].setCurrentIndex(-1)
            self.band_comboboxes[i].activated.connect(self.test_enable_ok)
            self.band_labels.append(QLabel(output_bands_names[i]))
        # mask combobox label and checkbox
        self.mask_combobox = QComboBox()
        self.mask_combobox.addItems(input_bands)
        if len(input_bands) == nb_output_bands + 1:
            self.mask_combobox.setCurrentIndex(nb_output_bands)
        else:
            self.mask_combobox.setCurrentIndex(-1)
        self.mask_combobox.activated.connect(self.test_enable_ok)
        self.mask_combobox.setDisabled(True)
        self.mask_label = QLabel("Mask")
        self.mask_label.setDisabled(True)
        self.mask_checkbox = QCheckBox()
        self.mask_checkbox.stateChanged.connect(self.mask_combobox.setEnabled)
        self.mask_checkbox.stateChanged.connect(self.mask_label.setEnabled)
        self.mask_checkbox.stateChanged.connect(self.test_enable_ok)
        self.mask_checkbox.setChecked(False)
        # input layout
        self.input_layout = QGridLayout()
        for i in range(nb_output_bands):
            self.input_layout.addWidget(self.band_labels[i], i, 1)
            self.input_layout.addWidget(self.band_comboboxes[i], i, 2)
        self.input_layout.addWidget(self.mask_checkbox, nb_output_bands, 0)
        self.input_layout.addWidget(self.mask_label, nb_output_bands, 1)
        self.input_layout.addWidget(self.mask_combobox, nb_output_bands, 2)
        # buttons
        self.button_box = QDialogButtonBox()
        self.cancel_button = self.button_box.addButton(QDialogButtonBox.StandardButton.Cancel)
        self.cancel_button.clicked.connect(self.reject)
        self.ok_button = self.button_box.addButton(QDialogButtonBox.StandardButton.Ok)
        self.ok_button.clicked.connect(self.accept)
        # main layout
        self.main_layout = QVBoxLayout()
        self.main_layout.addLayout(self.input_layout)
        self.main_layout.addWidget(self.button_box)
        self.setLayout(self.main_layout)
        self.test_enable_ok()

    @Slot()
    def test_enable_ok(self) -> None:
        enable_ok = True
        for combobox in self.band_comboboxes:
            if combobox.currentIndex() == -1:
                enable_ok = False
        if self.mask_checkbox.checkState() == Qt.CheckState.Checked and self.mask_combobox.currentIndex() == -1:
            enable_ok = False
        self.ok_button.setEnabled(enable_ok)



# -*- coding: utf-8 -*-

"""
Quaternion to have nice rotations in 3D.

Copyright (c) 2010, Renaud Blanch <rndblnch at gmail dot com>
Licence: GPLv3 or higher <http://www.gnu.org/licenses/gpl.html>
"""

# imports ####################################################################

from math import (cos as _cos, sin as _sin, atan2 as _atan2, asin as _asin,
                  sqrt as _sqrt)

from . import vector as _v


# quaternion #################################################################

def quaternion(theta=0, u=(1., 0., 0.)):
    w = _cos(theta/2.)
    x, y, z = (ui*_sin(theta/2.) for ui in u)
    return w, (x, y, z)


def arcball(x, y):
    h2 = x*x+y*y
    if h2 > 1.:
        h = _sqrt(h2)
        v = x/h, y/h, 0.
    else:
        v = x, y, _sqrt(1.-h2)
    return 0., v


def mul(P, Q):
    w1, v1 = P
    w2, v2 = Q
    return (w1*w2 - _v.dot(v1, v2),
            _v.sum(_v.mul(w1, v2), _v.mul(w2, v1), _v.cross(v1, v2)))


def product(P=quaternion(), *Qs):
    for Q in Qs:
        P = mul(P, Q)
    return P


def matrix(Q):
    w, (x, y, z) = Q
    return [[1.-2.*(y*y+z*z), 2.*(x*y+w*z),    2.*(x*z-w*y),    0.],
            [2.*(x*y-w*z),    1.-2.*(x*x+z*z), 2.*(y*z+w*x),    0.],
            [2.*(x*z+w*y),    2.*(y*z-w*x),    1.-2.*(x*x+y*y), 0.],
            [0.,              0.,              0.,              1.]]


def euler_angles(Q):
    w, (x, y, z) = Q
    phi = _atan2(2.*(w*x+y*z), 1.-2.*(x*x+y*y))
    theta = _asin(2*(w*y-x*z))
    psi = _atan2(2.*(w*z+x*y), 1.-2.*(y*y+z*z))
    return phi, theta, psi


def antipod(Q):
    w, v = Q
    return -w, _v.mul(-1., v)


def inverse(Q):
    w, v = Q
    return w, _v.mul(-1., v)


def rotate(Q, v):
    _, v = product(Q, (0, v), inverse(Q))
    return v


def theta_u(Q):
    w, v = Q
    norm = _v.norm(v)
    theta = 2.*_atan2(norm, w)
    u = _v.mul(1./norm, v)
    return theta, u


def power(Q, a):
    theta, u = theta_u(Q)
    return quaternion(a*theta, u)


def slerp(Q0, Q1, a):
    return mul(power(mul(Q1, inverse(Q0)), a), Q0)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" custom_widgets

This modules contains custom widgets.

Contains classes:
    * Toggle - round toggle button switching between two positions
    * AnimatedToggle - animates the Toggle switch
    * FileInfoWidget - table-like widget filed with metadata from file

modified from qt-widgets: github.com/pythonguis/python-qtwidgets

"""

from typing import Optional

import math

from PySide6.QtCore import (
    Qt, QSize, QPoint, QPointF, QRectF, QTimer, QRect,
    QEasingCurve, QPropertyAnimation, QSequentialAnimationGroup, QAbstractAnimation
)

from PySide6.QtCore import Slot, Property

from PySide6.QtWidgets import (
    QWidget, QCheckBox, QStyle, QProxyStyle, QStyleOption, QTabBar, QDockWidget, QMainWindow
)

from PySide6.QtGui import QColor, QBrush, QPaintEvent, QPen, QPainter, QIcon


class IconDockStyle(QProxyStyle):

    # adapted from https://stackoverflow.com/a/3482795

    def __init__(self, icon: QIcon, style: Optional[QStyle] = None):
        super().__init__(style)
        self.icon = icon

    def drawControl(self, element: QStyle.ControlElement, option: QStyleOption, painter: QPainter,
                    widget: Optional[QWidget] = None) -> None:
        if element == QStyle.ControlElement.CE_DockWidgetTitle:
            # width of the icon
            width: int = self.pixelMetric(QStyle.PixelMetric.PM_TabBarIconSize)
            # margin of title from frame
            margin: int = self.baseStyle().pixelMetric(QStyle.PixelMetric.PM_DockWidgetTitleMargin)
            icon_point = QPoint(margin + option.rect.left(),
                                margin + option.rect.center().y() - width//2)
            painter.drawPixmap(icon_point, self.icon.pixmap(width, width))
            option.rect = option.rect.adjusted(width, 0, 0, 0)
        self.baseStyle().drawControl(element, option, painter, widget)


def add_icon_to_tab(window: QMainWindow, tab_name: str, icon: QIcon):
    # adapted from https://stackoverflow.com/a/46623219
    def f(_):
        try:
            for tabbar in window.findChildren((QTabBar)):
                for i in range(tabbar.count()):
                    if tabbar.tabText(i) == tab_name:
                        tabbar.setTabIcon(i, icon)
        except RuntimeError:
            # window has been deleted
            pass
    return f


class IconDockWidget(QDockWidget):

    def __init__(self, name: str, parent: QWidget, icon: QIcon):
        assert isinstance(parent, QMainWindow)
        super().__init__(name, parent)
        self.setStyle(IconDockStyle(icon))
        self.visibilityChanged.connect(add_icon_to_tab(parent, name, icon))


class Toggle(QCheckBox):

    _transparent_pen = QPen(Qt.GlobalColor.transparent)
    _light_grey_pen = QPen(Qt.GlobalColor.lightGray)

    def __init__(self, parent=None, bar_color: QColor = QColor("gray"),
                 checked_color: QColor = QColor("white"), handle_color: QColor = QColor("white")):
        super().__init__(parent)
        # Save our properties on the object via self, so we can access them
        # later in the paintEvent.
        self._bar_brush = QBrush(bar_color)
        self._bar_checked_brush = QBrush(QColor(checked_color).lighter())
        self._handle_brush = QBrush(handle_color)
        self._handle_checked_brush = QBrush(QColor(checked_color))
        # Setup the rest of the widget.
        self.setContentsMargins(8, 0, 8, 0)
        self._handle_position = 0
        self.stateChanged.connect(self.handle_state_change)

    def sizeHint(self) -> QSize:
        return QSize(58, 45)

    def hitButton(self, pos: QPoint) -> bool:
        return self.contentsRect().contains(pos)

    def paintEvent(self, e: QPaintEvent) -> None:
        # pylint: disable=missing-function-docstring, invalid-name, unused-argument
        contRect = self.contentsRect()
        handleRadius = round(0.24 * contRect.height())

        p = QPainter(self)
        p.setRenderHint(QPainter.RenderHint.Antialiasing)

        p.setPen(self._transparent_pen)
        barRect = QRectF(
            0, 0,
            contRect.width() - handleRadius, 0.40 * contRect.height()
        )
        barRect.moveCenter(contRect.center())
        rounding = barRect.height() / 2

        # the handle will move along this line
        trailLength = contRect.width() - 2 * handleRadius
        xPos = (
            contRect.x() + handleRadius + trailLength * self._handle_position
        )

        if self.isChecked():
            p.setBrush(self._bar_checked_brush)
            p.drawRoundedRect(barRect, rounding, rounding)
            p.setPen(self._light_grey_pen)

            p.setBrush(self._handle_checked_brush)

        else:
            p.setBrush(self._bar_brush)
            p.drawRoundedRect(barRect, rounding, rounding)
            p.setPen(self._light_grey_pen)
            p.setBrush(self._handle_brush)

        p.drawEllipse(
            QPointF(xPos, barRect.center().y()),
            handleRadius, handleRadius)

        p.end()

    @Slot(int)
    def handle_state_change(self, value: int) -> None:
        self._handle_position = 1 if value else 0

    @Property(float)
    def handle_position(self) -> float:
        return self._handle_position

    @handle_position.setter
    def handle_position(self, pos):
        """change the property
        we need to trigger QWidget.update() method, either by:
            1- calling it here [ what we're doing ].
            2- connecting the QPropertyAnimation.valueChanged() signal to it.
        """
        self._handle_position = pos
        self.update()

    @Property(float)
    def pulse_radius(self):
        return self._pulse_radius

    @pulse_radius.setter
    def pulse_radius(self, pos):
        self._pulse_radius = pos
        self.update()


class AnimatedToggle(Toggle):

    _transparent_pen = QPen(Qt.GlobalColor.transparent)
    _light_grey_pen = QPen(Qt.GlobalColor.lightGray)

    def __init__(self,
                 *args,
                 pulse_unchecked_color="#44999999",
                 pulse_checked_color="#44999999",
                 **kwargs):

        self._pulse_radius: float = 0

        super().__init__(*args, **kwargs)

        self.animation = QPropertyAnimation(self, b"handle_position", self)
        self.animation.setEasingCurve(QEasingCurve.InOutCubic)
        self.animation.setDuration(200)  # time in ms

        self.pulse_anim = QPropertyAnimation(self, b"pulse_radius", self)
        self.pulse_anim.setDuration(350)  # time in ms
        self.pulse_anim.setStartValue(10)
        self.pulse_anim.setEndValue(20)

        self.animations_group = QSequentialAnimationGroup()
        self.animations_group.addAnimation(self.animation)
        self.animations_group.addAnimation(self.pulse_anim)

        self._pulse_unchecked_animation = QBrush(QColor(pulse_unchecked_color))
        self._pulse_checked_animation = QBrush(QColor(pulse_checked_color))

    @Slot(int)
    def handle_state_change(self, value):
        self.animations_group.stop()
        if value:
            self.animation.setEndValue(1)
        else:
            self.animation.setEndValue(0)
        self.animations_group.start()

    def paintEvent(self, e: QPaintEvent):

        contRect = self.contentsRect()
        handleRadius = int(round(0.24 * contRect.height()))

        p = QPainter(self)
        p.setRenderHint(QPainter.RenderHint.Antialiasing)

        p.setPen(self._transparent_pen)
        barRect = QRectF(
            0, 0,
            contRect.width() - handleRadius, handleRadius*2
            # 0.25 * contRect.height()
        )
        barRect.moveCenter(contRect.center())
        rounding = barRect.height() / 2

        # the handle will move along this line
        trailLength = contRect.width() - 2 * handleRadius

        xPos = (
            contRect.x() + handleRadius + trailLength * self._handle_position
        )

        if self.pulse_anim.state() == QAbstractAnimation.State.Running:
            p.setBrush(
                self._pulse_checked_animation if
                self.isChecked() else self._pulse_unchecked_animation)
            p.drawEllipse(QPointF(xPos, barRect.center().y()),
                          self._pulse_radius, self._pulse_radius)

        if self.isChecked():
            p.setBrush(self._bar_checked_brush)
            p.setPen(self._light_grey_pen)
            p.drawRoundedRect(barRect, rounding, rounding)
            p.setBrush(self._handle_checked_brush)

        else:
            p.setBrush(self._bar_brush)
            p.setPen(self._light_grey_pen)
            p.drawRoundedRect(barRect, rounding, rounding)
            p.setBrush(self._handle_brush)

        p.drawEllipse(
            QPointF(xPos, barRect.center().y()),
            handleRadius, handleRadius)

        p.setBrush(Qt.GlobalColor.black)
        p.drawPie(int(xPos)-handleRadius,
                  int(barRect.center().y())-handleRadius,
                  handleRadius*2,
                  handleRadius*2,
                  90*16,
                  180*16)

        # text under icon:
        # p.setPen(Qt.black)
        # small_font = p.font()
        # small_font.setPointSize(p.font().pointSize()//2)
        # p.setFont(small_font)
        # p.drawText(contRect,
        #            (Qt.AlignCenter | Qt.AlignBottom),
        #            'background \ncolor')

        p.end()


# from https://github.com/z3ntu/QtWaitingSpinner

class QtWaitingSpinner(QWidget):

    def __init__(self, parent, centerOnParent=True, disableParentWhenSpinning=True, modality=Qt.NonModal):
        super().__init__(parent)

        self._centerOnParent = centerOnParent
        self._disableParentWhenSpinning = disableParentWhenSpinning

        # WAS IN initialize()
        self._color = QColor(Qt.black)
        self._roundness = 50.0
        self._minimumTrailOpacity = 30
        self._trailFadePercentage = 60.0
        self._revolutionsPerSecond = 1
        self._numberOfLines = 12
        self._lineLength = 10
        self._lineWidth = 2
        self._innerRadius = 10
        self._currentCounter = 0
        self._isSpinning = False

        self._timer = QTimer(self)
        self._timer.timeout.connect(self.rotate)
        self.updateSize()
        self.updateTimer()
        self.hide()
        # END initialize()

        self.setWindowModality(modality)
        self.setAttribute(Qt.WA_TranslucentBackground)

    def paintEvent(self, event: Optional[QPaintEvent]):
        self.updatePosition()
        painter = QPainter(self)
        painter.fillRect(self.rect(), Qt.GlobalColor.transparent)
        painter.setRenderHint(QPainter.RenderHint.Antialiasing, True)

        if self._currentCounter >= self._numberOfLines:
            self._currentCounter = 0

        painter.setPen(Qt.PenStyle.NoPen)
        for i in range(0, self._numberOfLines):
            painter.save()
            painter.translate(self._innerRadius + self._lineLength,
                              self._innerRadius + self._lineLength)
            rotateAngle = float(360 * i) / float(self._numberOfLines)
            painter.rotate(rotateAngle)
            painter.translate(self._innerRadius, 0)
            distance = self.lineCountDistanceFromPrimary(
                i, self._currentCounter, self._numberOfLines)
            color = self.currentLineColor(distance, self._numberOfLines, self._trailFadePercentage,
                                          self._minimumTrailOpacity, self._color)
            painter.setBrush(color)
            rect = QRect(0, int(-self._lineWidth / 2), int(self._lineLength), int(self._lineWidth))
            painter.drawRoundedRect(rect, self._roundness, self._roundness,
                                    Qt.SizeMode.RelativeSize)
            painter.restore()

    def start(self):
        self.updatePosition()
        self._isSpinning = True
        self.show()
        if self.parentWidget and self._disableParentWhenSpinning:
            self.parentWidget().setEnabled(False)
        if not self._timer.isActive():
            self._timer.start()
            self._currentCounter = 0

    def stop(self):
        self._isSpinning = False
        self.hide()

        if self.parentWidget() and self._disableParentWhenSpinning:
            self.parentWidget().setEnabled(True)

        if self._timer.isActive():
            self._timer.stop()
            self._currentCounter = 0

    def setNumberOfLines(self, lines):
        self._numberOfLines = lines
        self._currentCounter = 0
        self.updateTimer()

    def setLineLength(self, length):
        self._lineLength = length
        self.updateSize()

    def setLineWidth(self, width):
        self._lineWidth = width
        self.updateSize()

    def setInnerRadius(self, radius):
        self._innerRadius = radius
        self.updateSize()

    def color(self):
        return self._color

    def roundness(self):
        return self._roundness

    def minimumTrailOpacity(self):
        return self._minimumTrailOpacity

    def trailFadePercentage(self):
        return self._trailFadePercentage

    def revolutionsPersSecond(self):
        return self._revolutionsPerSecond

    def numberOfLines(self):
        return self._numberOfLines

    def lineLength(self):
        return self._lineLength

    def lineWidth(self):
        return self._lineWidth

    def innerRadius(self):
        return self._innerRadius

    def isSpinning(self):
        return self._isSpinning

    def setRoundness(self, roundness):
        self._roundness = max(0.0, min(100.0, roundness))

    def setColor(self, color=Qt.black):
        self._color = QColor(color)

    def setRevolutionsPerSecond(self, revolutionsPerSecond):
        self._revolutionsPerSecond = revolutionsPerSecond
        self.updateTimer()

    def setTrailFadePercentage(self, trail):
        self._trailFadePercentage = trail

    def setMinimumTrailOpacity(self, minimumTrailOpacity):
        self._minimumTrailOpacity = minimumTrailOpacity

    def rotate(self):
        self._currentCounter += 1
        if self._currentCounter >= self._numberOfLines:
            self._currentCounter = 0
        self.update()

    def updateSize(self):
        size = int((self._innerRadius + self._lineLength) * 2)
        self.setFixedSize(size, size)

    def updateTimer(self):
        self._timer.setInterval(int(1000 / (self._numberOfLines * self._revolutionsPerSecond)))

    def updatePosition(self):
        if self.parentWidget() and self._centerOnParent:
            self.move(int(self.parentWidget().width() / 2 - self.width() / 2),
                      int(self.parentWidget().height() / 2 - self.height() / 2))

    def lineCountDistanceFromPrimary(self, current, primary, totalNrOfLines):
        distance = primary - current
        if distance < 0:
            distance += totalNrOfLines
        return distance

    def currentLineColor(self, countDistance, totalNrOfLines, trailFadePerc, minOpacity, colorinput):
        color = QColor(colorinput)
        if countDistance == 0:
            return color
        minAlphaF = minOpacity / 100.0
        distanceThreshold = int(math.ceil((totalNrOfLines - 1) * trailFadePerc / 100.0))
        if countDistance > distanceThreshold:
            color.setAlphaF(minAlphaF)
        else:
            alphaDiff = color.alphaF() - minAlphaF
            gradient = alphaDiff / float(distanceThreshold + 1)
            resultAlpha = color.alphaF() - gradient * countDistance
            # If alpha is out of bounds, clip it.
            resultAlpha = min(1.0, max(0.0, resultAlpha))
            color.setAlphaF(resultAlpha)
        return color



# -*- coding: utf-8 -*-
""" PlotView

This module contains classes pertaining to the display of data in plot views.
Data selected by the user is displayed in several plots.

Works with the PlotModel module, as in a Model/View architecture.
"""

from typing import Any, Optional, Iterable, Union, overload

import datetime

import pyqtgraph as pg

from pyqtgraph.GraphicsScene import exportDialog

from PySide6.QtCore import (
    Qt, QObject, Signal, Slot, QAbstractProxyModel, QAbstractItemModel, QModelIndex
)

from PySide6.QtWidgets import QWidget, QVBoxLayout, QToolBar, QPushButton

from insarviz.custom_widgets import AnimatedToggle

from insarviz.plot.PlotModel import PlotModel


# PlotWidget classes ###############################################################################

class AbstractPlotWidget(pg.PlotWidget):
    """
    Based on pyqtgraph's GraphicsLayoutWidget
    """

    sigcurveSelected = Signal()

    default_padding = 0.05
    max_padding = 0.15

    # style colors
    DARK_THEME: bool = False
    LIGHT_THEME: bool = True

    def __init__(self, plot_model: PlotModel):
        super().__init__()
        self.plot_model: PlotModel = plot_model
        self.tooltip_text: dict[str, str] = {}
        self.plotItem.setLabel('left', "LOS Displacement")
        self.plotItem.getAxis('left').enableAutoSIPrefix(False)
        self.plotItem.getViewBox().setDefaultPadding(self.default_padding)
        plot_model.loader.data_units_loaded.connect(self.set_data_units)

    @Slot(object, object, object)
    def on_points_hover(self, curve: pg.PlotDataItem, points: list, _) -> None:
        if len(points) == 0:
            if curve.name() in self.tooltip_text:
                del self.tooltip_text[curve.name()]
        else:
            def x_format(x):
                if isinstance(self.plotItem.getAxis('bottom'), pg.DateAxisItem):
                    return datetime.datetime.fromtimestamp(x).strftime('%Y-%m-%d')
                return int(x)

            def y_format(y):
                units = f" {self.plot_model.loader.units}"
                if units == ' Undefined units':
                    units = ''
                return f"{y:.3f}{units}"

            self.tooltip_text[curve.name()] = curve.name() + "\n\n" + "\n\n".join(
                [f"x: {x_format(p.pos().x())}\ny: {y_format(p.pos().y())}" for p in points])
        text = "\n\n".join([v for v in self.tooltip_text.values()])
        self.plotItem.getViewBox().setToolTip(text)

    @Slot(str)
    def set_data_units(self, units: str):
        """
        Set the unit string of the Y axis "LOS Displacement {units}"
        """
        self.plotItem.setLabel('left', "LOS Displacement", units=units)

    @Slot(bool)
    def set_style(self, style: bool) -> None:
        """
        Set the plot theme (light or dark). Called when clicking on theme toggle,
        alternate between dark and light.

        Parameters
        ----------
        style : bool
            The state of the toggle, compared to self.DARK_THEME and self.LIGHT_THEME
        """
        (ax1, ax2) = (self.plotItem.getAxis('bottom'), self.plotItem.getAxis('left'))
        if style == self.DARK_THEME:
            # black background
            self.setBackground('k')
            self.plotItem.setTitle(self.plotItem.titleLabel.text, color=.5)
            ax1.setPen('w')
            ax2.setPen('w')
            ax1.setTextPen('w')
            ax2.setTextPen('w')
        elif style == self.LIGHT_THEME:
            # white background
            self.setBackground('w')
            self.plotItem.setTitle(self.plotItem.titleLabel.text, color='k')
            ax1.setPen('k')
            ax2.setPen('k')
            ax1.setTextPen('k')
            ax2.setTextPen('k')
        # update view: for some reason, just 'show' is not enough:
        self.repaint()


# PlotWindow classes ###############################################################################

class AbstractPlotWindow(QWidget):
    """
    Window to display a PlotWidget
    """

    def __init__(self, plot_model: PlotModel, plot_widget: AbstractPlotWidget):
        super().__init__()
        self.plot_model = plot_model
        self.plot_widget: AbstractPlotWidget = plot_widget
        self.plot_widget.set_style(AbstractPlotWidget.LIGHT_THEME)
        # toolbar
        self.toolbar: QToolBar = QToolBar(self)
        self.theme_switch: AnimatedToggle = AnimatedToggle()
        self.theme_switch.setChecked(True)
        self.theme_switch.toggled.connect(self.plot_widget.set_style)
        self.theme_switch.setToolTip('Theme switch (light/dark background)')
        self.toolbar.addWidget(self.theme_switch)
        self.autorange_button: QPushButton = QPushButton("Autorange", self)
        self.autorange_button.setToolTip("Automatically adjust axis range")
        self.autorange_button.clicked.connect(self.autorange)
        self.toolbar.addWidget(self.autorange_button)
        self.export_button: QPushButton = QPushButton("Export...", self)
        self.export_button.clicked.connect(self.export)
        # export dialog
        self.export_dialog = exportDialog.ExportDialog(self.plot_widget.plotItem.scene())
        # layout
        layout = QVBoxLayout()
        layout.setMenuBar(self.toolbar)
        layout.setContentsMargins(5, 5, 5, 5)
        layout.setSpacing(5)
        self.setLayout(layout)

    def autorange(self) -> None:
        self.plot_widget.getPlotItem().getViewBox().autoRange()
        self.plot_widget.getPlotItem().autoBtnClicked()

    def export(self) -> None:
        self.export_dialog.show(self.plot_widget.plotItem)

# ProxyNoItemModel class ###########################################################################


class ProxyNoItemModel(QAbstractProxyModel):

    # add a 0th row to a source model, with noitem_text and that can be selected

    def __init__(self, source_model: QAbstractItemModel, noitem_text: str):
        super().__init__()
        self.setSourceModel(source_model)
        self.noitem_string = noitem_text

    # methods reimplemented from QAbstractProxyModel

    def setSourceModel(self, model: QAbstractItemModel) -> None:
        # pylint: disable=missing-function-docstring, invalid-name
        if self.sourceModel() is not None:
            self.sourceModel().rowsAboutToBeInserted.disconnect(self.source_rows_about_inserted)
            self.sourceModel().rowsInserted.disconnect(self.source_rows_inserted)
            self.sourceModel().rowsAboutToBeRemoved.disconnect(self.source_rows_about_removed)
            self.sourceModel().rowsRemoved.disconnect(self.source_rows_removed)
            self.sourceModel().dataChanged.disconnect(self.source_data_changed)
        super().setSourceModel(model)
        self.sourceModel().rowsAboutToBeInserted.connect(self.source_rows_about_inserted)
        self.sourceModel().rowsInserted.connect(self.source_rows_inserted)
        self.sourceModel().rowsAboutToBeRemoved.connect(self.source_rows_about_removed)
        self.sourceModel().rowsRemoved.connect(self.source_rows_removed)
        self.sourceModel().dataChanged.connect(self.source_data_changed)

    def rowCount(self, parent: QModelIndex = QModelIndex()) -> int:
        # pylint: disable=missing-function-docstring, invalid-name
        if parent == QModelIndex():
            return self.sourceModel().rowCount(QModelIndex()) + 1
        return 0

    def columnCount(self, parent: QModelIndex = QModelIndex()) -> int:
        # pylint: disable=missing-function-docstring, invalid-name, unused-argument
        return 1

    def index(self, row: int, column: int, parent: QModelIndex = QModelIndex()) -> QModelIndex:
        # pylint: disable=missing-function-docstring, invalid-name
        if parent.isValid() or self.rowCount(parent) <= row or self.columnCount(parent) <= column:
            return QModelIndex()
        if row > 0:
            pointer = self.sourceModel().index(row-1, column, QModelIndex()).internalPointer()
        else:
            pointer = None
        return self.createIndex(row, column, pointer)

    # see https://mypy.readthedocs.io/en/stable/more_types.html#function-overloading
    @overload
    def parent(self, child: QModelIndex) -> QModelIndex: ...

    @overload
    def parent(self) -> QObject: ...

    def parent(self, child: Optional[QModelIndex] = None) -> Union[QObject, QModelIndex]:
        # pylint: disable=missing-function-docstring
        if child is None:
            return super().parent()
        return QModelIndex()

    def data(self, index: QModelIndex, role: int = Qt.ItemDataRole.DisplayRole) -> Any:
        # pylint: disable=missing-function-docstring
        if role == Qt.ItemDataRole.CheckStateRole:
            # no checkbox
            return None
        if index.isValid() and index.row() == 0:
            # Noitem item
            if role == Qt.ItemDataRole.DisplayRole:
                return self.noitem_string
            return None
        return super().data(index, role)

    def flags(self, index: QModelIndex) -> Qt.ItemFlags:
        # pylint: disable=missing-function-docstring
        if index.isValid() and index.row() == 0:
            # Noitem item
            return Qt.ItemFlag.ItemIsSelectable | Qt.ItemFlag.ItemIsEnabled
        return super().flags(index)

    def mapFromSource(self, sourceIndex: QModelIndex) -> QModelIndex:
        # pylint: disable=missing-function-docstring, invalid-name
        if not sourceIndex.isValid():
            return QModelIndex()
        return self.index(sourceIndex.row()+1, sourceIndex.column(), QModelIndex())

    def mapToSource(self, proxyIndex: QModelIndex) -> QModelIndex:
        # pylint: disable=missing-function-docstring, invalid-name
        if not proxyIndex.isValid():
            return QModelIndex()
        if proxyIndex.row() == 0:
            return QModelIndex()
        return self.sourceModel().index(proxyIndex.row()-1, proxyIndex.column(), QModelIndex())

    # custom methods

    @Slot(QModelIndex, int, int)
    def source_rows_about_inserted(self, parent: QModelIndex, first: int, last: int) -> None:
        # pylint: disable=missing-function-docstring
        self.beginInsertRows(self.mapFromSource(parent), first + 1, last + 1)

    @Slot(QModelIndex, int, int)
    def source_rows_inserted(self, parent: QModelIndex, first: int, last: int) -> None:
        # pylint: disable=missing-function-docstring, disable=unused-argument
        self.endInsertRows()

    @Slot(QModelIndex, int, int)
    def source_rows_about_removed(self, parent: QModelIndex, first: int, last: int) -> None:
        # pylint: disable=missing-function-docstring
        self.beginRemoveRows(self.mapFromSource(parent), first + 1, last + 1)

    @Slot(QModelIndex, int, int)
    def source_rows_removed(self, parent: QModelIndex, first: int, last: int) -> None:
        # pylint: disable=unused-argument, missing-function-docstring
        self.endRemoveRows()

    #  TODO cannot find correct Slot signature:
    # @Slot(QModelIndex, QModelIndex, list) not working
    def source_data_changed(self, top_left: QModelIndex, bottom_right: QModelIndex,
                            roles: Iterable[int]) -> None:
        # pylint: disable=missing-function-docstring
        self.dataChanged.emit(self.mapFromSource(top_left), self.mapFromSource(bottom_right), roles)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from typing import Optional, Union

import datetime

import numpy as np

from PySide6.QtCore import Qt, QPointF, Slot, Signal, QSignalBlocker, QEvent

from PySide6.QtGui import QGuiApplication, QKeyEvent

from PySide6.QtWidgets import (
    QWidget, QStyle, QProxyStyle, QStyleOption, QStyleHintReturn, QSlider, QLabel,
    QSpinBox, QHBoxLayout, QVBoxLayout, QGroupBox
)

import pyqtgraph as pg

from insarviz.utils import get_nearest

from insarviz.Loader import Loader


class SliderLeftClickStyle(QProxyStyle):
    """
    ProxyStyle overriding the mouse button used to move a slider on click (all buttons now move
    the slider to the position under the mouse)

    see https://stackoverflow.com/a/26281608
    """

    def styleHint(self, hint: QStyle.StyleHint, option: Optional[QStyleOption] = None,
                  widget: Optional[QWidget] = None,
                  returnData: Optional[QStyleHintReturn] = None) -> int:
        if hint == QStyle.StyleHint.SH_Slider_AbsoluteSetButtons:
            return (Qt.MouseButton.LeftButton | Qt.MouseButton.MiddleButton |
                    Qt.MouseButton.RightButton).value
        return super().styleHint(hint, option, widget, returnData)


class TimelineTick(pg.InfiniteLine):

    tick_pen = pg.functions.mkPen("grey", width=2)
    hover_pen = pg.functions.mkPen("r", width=3)

    def __init__(self, pos: float, tooltip: str = ""):
        super().__init__(pos=pos, pen=TimelineTick.tick_pen,
                         hoverPen=TimelineTick.hover_pen, angle=90, movable=True)
        self.setToolTip(tooltip)

    def mouseDragEvent(self, ev) -> None:
        pass

    def mouseClickEvent(self, ev) -> None:
        self.sigClicked.emit(self, ev)


class TimelineHandle(pg.InfiniteLine):

    handle_pen = pg.functions.mkPen("blue", width=3)
    dragged_pen = pg.functions.mkPen("cyan", width=3)
    hover_pen = pg.functions.mkPen("red", width=3)

    def __init__(self, loader: Loader):
        self.loader = loader
        super().__init__(pen=TimelineHandle.handle_pen, hoverPen=TimelineHandle.hover_pen,
                         angle=90, movable=True)
        self.addMarker('<|>')
        self.setZValue(1)
        self.setCursor(Qt.CursorShape.OpenHandCursor)
        self.sigDragged.connect(lambda x: x.setPen(TimelineHandle.dragged_pen))
        self.sigPositionChangeFinished.connect(lambda x: x.setPen(TimelineHandle.handle_pen))

    def setPos(self, pos: Union[float, QPointF]) -> None:
        if isinstance(pos, QPointF):
            pos = pos.x()
        elif isinstance(pos, (list, tuple, np.ndarray)):
            pos = pos[0]
        if self.loader.timestamps is not None and pos not in self.loader.timestamps:
            pos, _ = get_nearest(self.loader.timestamps, pos)
        super().setPos(pos)
        self.setToolTip(str(datetime.date.fromtimestamp(pos)))

    def mouseDragEvent(self, ev) -> None:
        super().mouseDragEvent(ev)
        if ev.isAccepted():
            if ev.isStart():
                QGuiApplication.instance().setOverrideCursor(Qt.CursorShape.DragMoveCursor)
            elif ev.isFinish():
                QGuiApplication.instance().restoreOverrideCursor()


class TimelineSlider(pg.PlotWidget):

    value_changed = Signal(int)

    def __init__(self, loader: Loader, parent: Optional[QWidget] = None):
        self.date_axis = pg.DateAxisItem()
        self.date_axis.enableAutoSIPrefix(False)
        super().__init__(parent=parent, axisItems={'bottom': self.date_axis}, enableMenu=False)
        self.loader = loader
        self.setMouseEnabled(y=False)
        plot_item = self.getPlotItem()
        plot_item.hideAxis('left')
        plot_item.hideButtons()
        self.ticks: list[TimelineTick] = []
        self.handle: Optional[TimelineHandle] = None
        self.setMaximumHeight(48)
        self.hide()

    def clear(self) -> None:
        if self.handle is not None:
            self.handle.sigPositionChanged.disconnect(self.on_handle_position_changed)
        for tick in self.ticks:
            tick.sigClicked.disconnect(self.on_tick_click)
        super().clear()

    @Slot(int)
    def on_data_loaded(self) -> None:
        assert self.loader.timestamps is not None
        # set the min max range of the x axis
        timedelta = self.loader.timestamps[-1] - self.loader.timestamps[0]
        xmin = self.loader.timestamps[0] - 0.1*timedelta
        xmax = self.loader.timestamps[-1] + 0.1*timedelta
        self.setLimits(xMin=xmin, xMax=xmax)
        self.getPlotItem().getViewBox().setXRange(self.loader.timestamps[0],
                                                  self.loader.timestamps[-1], padding=0.1)
        # date marker
        for d, t in zip(self.loader.dates, self.loader.timestamps):
            assert isinstance(d, datetime.datetime)
            tooltip = str(d.date())
            tick = TimelineTick(t, tooltip)
            tick.sigClicked.connect(self.on_tick_click)
            self.addItem(tick)
            self.ticks.append(tick)
        self.handle = TimelineHandle(self.loader)
        self.handle.sigPositionChanged.connect(self.on_handle_position_changed)
        self.addItem(self.handle)

    @Slot(int)
    def set_value(self, band_index: int) -> None:
        assert self.loader.dataset is not None and self.loader.timestamps is not None
        assert self.handle is not None
        self.handle.setPos(self.loader.timestamps[self.loader.dataset.indexes.index(band_index)])

    @Slot(object, object)
    def on_tick_click(self, tick: TimelineTick, ev) -> None:
        if ev.button() == Qt.MouseButton.LeftButton:
            assert self.handle is not None
            self.handle.setPos(tick.getXPos())

    @Slot(object)
    def on_handle_position_changed(self, _: TimelineHandle) -> None:
        assert self.loader.dataset is not None and self.loader.timestamps is not None
        assert self.handle is not None
        assert self.handle.getXPos() in self.loader.timestamps
        index = self.loader.dataset.indexes[
            np.where(self.loader.timestamps == self.handle.getXPos())[0][0]]
        self.value_changed.emit(index)


class BandSlider(QWidget):

    value_changed = Signal(int)
    reference_band_changed = Signal(object)

    def __init__(self, loader: Loader, parent: Optional[QWidget] = None):
        super().__init__(parent)
        self.loader = loader
        self.loader.reference_band_index_changed.connect(self.set_reference_band_index)
        self.reference_band_changed.connect(self.loader.set_reference_band)
        # label
        self.date_label = QLabel("")
        self.date_label.setAlignment(Qt.AlignmentFlag.AlignTop | Qt.AlignmentFlag.AlignLeft)
        self.date_label.setMaximumHeight(20)
        # reference band spin box
        self.reference_band_groupbox = QGroupBox("Reference Band")
        self.reference_band_groupbox.setCheckable(True)
        self.reference_band_groupbox.setChecked(False)
        self.reference_band_groupbox.clicked.connect(self.on_reference_band_groupbox_clicked)
        self.reference_band_layout = QHBoxLayout()
        self.reference_band_groupbox.setLayout(self.reference_band_layout)
        self.reference_band_setter = mySpinBox()
        self.reference_band_setter.setRange(0, 0)
        self.reference_band_setter.editingFinished.connect(
            self.on_reference_band_setter_editingfinished)
        self.reference_band_layout.addWidget(self.reference_band_setter)
        # spin box
        self.band_setter = QSpinBox()
        self.band_setter.setRange(1, 1)
        self.band_setter.setKeyboardTracking(False)
        # slider
        self.slider = QSlider(Qt.Orientation.Horizontal)
        self.slider.setStyle(SliderLeftClickStyle())
        self.slider.setMinimum(1)
        self.slider.setMaximum(1)
        self.slider.setTickPosition(QSlider.TickPosition.TicksBelow)
        self.slider.setTickInterval(1)
        # allow arrow keys press to change date
        self.slider.setFocusPolicy(Qt.FocusPolicy.StrongFocus)
        # timeline slider
        self.timeline_slider = TimelineSlider(self.loader)
        # signals and slots
        self.band_setter.valueChanged.connect(self.set_value)
        self.slider.valueChanged.connect(self.set_value)
        self.timeline_slider.value_changed.connect(self.set_value)
        # layout
        self.layoutH = QHBoxLayout()
        self.layoutH.addWidget(self.date_label)
        self.layoutH.addWidget(self.band_setter)
        self.layoutH.addWidget(self.slider)
        self.layoutH.addWidget(self.reference_band_groupbox)
        self.layoutV = QVBoxLayout()
        self.layoutV.addLayout(self.layoutH)
        self.layoutV.addWidget(self.timeline_slider)
        self.setLayout(self.layoutV)
        # set disabled
        self.setDisabled(True)

    @Slot()
    def on_close(self) -> None:
        self.setDisabled(True)
        self.slider.blockSignals(True)
        self.slider.setValue(1)
        self.slider.blockSignals(False)
        with QSignalBlocker(self.band_setter):
            self.band_setter.setValue(1)
        with QSignalBlocker(self.reference_band_setter):
            self.reference_band_setter.setValue(1)
        with QSignalBlocker(self.reference_band_groupbox):
            self.reference_band_groupbox.setChecked(False)
        self.timeline_slider.clear()
        self.timeline_slider.hide()
        self.date_label.setText("File->Open to load data")
        self.date_label.setEnabled(True)

    @Slot(int)
    def on_data_loaded(self, _: int):
        self.set_range(1, len(self.loader))
        if isinstance(self.loader.dates[0], datetime.datetime):
            self.timeline_slider.on_data_loaded()
            self.timeline_slider.show()
        else:
            self.timeline_slider.hide()

    def set_range(self, start: int, end: int) -> None:
        self.slider.setMinimum(start)
        self.slider.setMaximum(end)
        self.band_setter.setRange(start, end)
        self.reference_band_setter.setRange(start, end)

    @Slot(int)
    def set_value(self, v: int) -> None:
        self.slider.setValue(v)
        self.band_setter.setValue(v)
        if not self.timeline_slider.isHidden():
            self.timeline_slider.set_value(v)
        assert self.loader.dataset is not None
        date = self.loader.dates[self.loader.dataset.indexes.index(v)]
        if isinstance(date, datetime.datetime):
            self.date_label.setText(f"{date.date()} - Band #")
        else:
            self.date_label.setText('Band #')
        self.value_changed.emit(v)

    @Slot(bool)
    def on_reference_band_groupbox_clicked(self, checked: bool) -> None:
        if not checked:
            self.reference_band_changed.emit(None)
        else:
            self.reference_band_changed.emit(self.reference_band_setter.value())

    @Slot(object)
    def set_reference_band_index(self, index: Optional[int]) -> None:
        if index is None:
            self.reference_band_groupbox.setChecked(False)
        else:
            index = int(index)
            self.reference_band_groupbox.setChecked(True)
            self.reference_band_setter.setValue(index)

    @Slot()
    def on_reference_band_setter_editingfinished(self) -> None:
        self.reference_band_changed.emit(self.reference_band_setter.value())


class mySpinBox(QSpinBox):

    def __init__(self, parent: QWidget = None):
        super().__init__(parent)

    def keyPressEvent(self, e: QKeyEvent) -> None:
        super().keyPressEvent(e)
        # accept "Enter" key press that would otherwise be forwarded to parent widget
        if e.key() in (Qt.Key.Key_Return, Qt.Key.Key_Enter):
            e.accept()

    def changeEvent(self, e: QEvent) -> None:
        # hide line edit when disabled
        if e.type() == QEvent.Type.EnabledChange:
            if not self.isEnabled():
                self.lineEdit().setText("None")
        super().changeEvent(e)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


# constants #################################################################

DATA_UNIT, COLORMAP_UNIT = range(2)  # texture unit use


# common shaders ############################################################

VERT_SHADER = r"""
    #version 330

    uniform mat4 model_matrix;
    uniform mat4 view_matrix;
    uniform mat4 projection_matrix;

    layout (location = 0) in vec3 vertex;
    layout (location = 1) in vec2 tex_coord;

    out vec2 tex_coord0;

    void main() {
        gl_Position = projection_matrix * view_matrix * model_matrix * vec4(vertex, 1.);
        tex_coord0 =  tex_coord;
    }
"""

ALPHA_SHADER = r"""
    #version 330

    uniform float alpha;

    vec4 apply_alpha(vec4 color) {
        return vec4(color.rgb, color.a*alpha);
    }
"""

COLORMAP_SHADER = r"""
    #version 330

    uniform sampler2D values;
    uniform sampler1D colormap;

    uniform float v0; // lower and
    uniform float v1; // upper bound of data values mapped to the colormap

    uniform float cutoff; // cutoff, to hide part of the layer

    vec4 apply_alpha(vec4 color);

    in vec2 tex_coord0;
    out vec4 frag_color;

    void main() {
        // get values (x) and mask (y)
        if(tex_coord0.x < cutoff) { discard; }
        vec4 t = texture(values, tex_coord0);
        vec2 v = t.xy;
        // if mask < 0.5 then more than half of values represented by pixel are nodata
        if(v.y < 0.5) { discard; }
        // get the color in the colormap 
        // "v.x/v.y" because v.y is the proportion of values that are not nodata
        vec4 l = texture(colormap, ((v.x/v.y)-v0)/(v1-v0));
        frag_color = apply_alpha(vec4(l.rgb, 1.));
    }
"""

IMAGE_RGB_SHADER = r"""
    #version 330

    vec4 apply_alpha(vec4 color);

    uniform sampler2D image;

    uniform float cutoff; // cutoff, to hide part of the layer

    in vec2 tex_coord0;
    out vec4 frag_color;

    void main() {
        if(tex_coord0.x < cutoff) { discard; }
        frag_color = apply_alpha(texture(image, tex_coord0));
    }
"""

# MiniMapView shaders for MapView viewport rect ####################################################

VIEWPORTRECT_VERT_SHADER = r"""
    #version 330

    uniform mat4 model_matrix;
    uniform mat4 view_matrix;
    uniform mat4 projection_matrix;

    layout (location = 0) in vec3 vertex;

    void main() {
        gl_Position = projection_matrix * view_matrix * model_matrix * vec4(vertex, 1.);
    }
"""

VIEWPORTRECT_FRAG_SHADER = r"""
    #version 330

    uniform vec4 rect_color;

    out vec4 frag_color;

    void main() {
        frag_color = rect_color;
    }
"""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

""" utils

This module contains miscellaneous utils methods:
* normalize_path - transform a path string into a pathlib.Path
* get_nearest - get index and value of closest value in array to target
value
* openUrl - open webpage of documentation
"""

# imports ###################################################################

from typing import Union

import pathlib

import numpy as np

from PySide6.QtWidgets import QMessageBox

from PySide6.QtCore import QUrl

from PySide6.QtGui import QDesktopServices

# utils #####################################################################


def normalize_path(path: str, project_path: str = "") -> pathlib.Path:
    """
    Transform a path string into a pathlib.Path.

    Parameters
    ----------
    path : str

    Returns
    -------
    pathlib.Path
    """
    # test if path is a correct absolute path to a file
    normalized_path = pathlib.Path(path)
    if normalized_path.exists():
        return normalized_path
    # test if path is a correct relative path to a file
    normalized_path = pathlib.Path(project_path).parent / pathlib.Path(path)
    if normalized_path.exists():
        return normalized_path
    # test if path is a correct absolute path to a file in Windows format
    normalized_path = pathlib.Path(pathlib.PureWindowsPath(path).as_posix())
    if normalized_path.exists():
        return normalized_path
    # test if path is a correct relative path to a file in Windows format
    normalized_path = pathlib.Path(pathlib.PureWindowsPath(
        project_path).parent.as_posix()) / pathlib.Path(pathlib.PureWindowsPath(path).as_posix())
    if normalized_path.exists():
        return normalized_path
    # else path may be an url
    return pathlib.Path(path)


def get_nearest(array, value):
    """
    Look into an array to find the nearest value to input value.
    Nearest is defined as minimal absolute difference between tested values.
    Returns array value and index of the nearest array value.

    Examples:
    ---------
    a = numpy.array((1, 5, 13, 7))
    get_nearest(a, 8)
    (7, 3)
    get_nearest(a, 1.23)
    (1, 0)

    Parameters
    ----------
    array : array
        Array where to look for the nearest value.
    value : int or float
        Value whose nearest value we want to find in the array.

    Returns
    -------
    array[idx], idx : tuple
        Nearest value found in array, index of nearest value found in array.

    """
    array = np.asarray(array)
    idx = (np.abs(array-value)).argmin()
    return array[idx], idx


def openUrl(self):
    """
    open url of documentation

    Returns
    -------
    None.

    """
    url = QUrl('https://deformvis.gricad-pages.univ-grenoble-alpes.fr/insarviz/index.html')  # noqa
    if not QDesktopServices.openUrl(url):
        QMessageBox.warning(self, 'Open Url', 'Could not open url')



# -*- coding: utf-8 -*-

from typing import Any, Union, Optional, TYPE_CHECKING

from PySide6.QtCore import Qt, QSize, QSignalBlocker, Slot

from PySide6.QtGui import QIcon

from PySide6.QtWidgets import (
    QWidget, QSpinBox, QDialog, QDialogButtonBox, QGridLayout, QLineEdit, QLabel, QCompleter,
    QGroupBox, QHBoxLayout
)

import rasterio

import numpy as np

import pathlib

import xyzservices as xyz

from insarviz.map.layers.Layer import OpenGLLayer

from insarviz.map.layers.TileLayer import TileLayer

if TYPE_CHECKING:
    from insarviz.map.MapModel import MapModel
    from insarviz.map.AbstractMapView import AbstractMapView


class XYZLayer(TileLayer):

    icon: QIcon = QIcon()
    kind: str = "XYZ layer"

    def __init__(self, name: str, model: "MapModel", map_widget: "AbstractMapView", provider: xyz.TileProvider):
        super().__init__(name, model, map_widget)
        if self.icon.isNull():
            # cannot be initialized in the class declaration because:
            # "QIcon needs a QGuiApplication instance before the icon is created." see QIcon doc
            XYZLayer.icon = QIcon('icons:WMS.svg')
        self.crs = rasterio.crs.CRS.from_epsg(4326)
        self.provider = provider
        if "max_zoom" not in self.provider:
            self.provider["max_zoom"] = 18
        # TODO test the tile size and that the server responds, otherwise raise Exception
        for i in range(0, self.provider.max_zoom+1):
            # transform the corners of the tilematrix into the main data set crs
            left, bottom, right, top = rasterio.warp.transform_bounds(self.crs, self.main_crs,
                                                                      -180, -85.0511, 180, 85.0511)
            # transform into coordinates in main dataset pixels
            left, top = ~self.main_transform * (left, top)
            right, bottom = ~self.main_transform * (right, bottom)
            tile_to_texture_pixelratioX = (right - left) / (256 * np.power(2, i))
            tile_to_texture_pixelratioY = (bottom - top) / (256 * np.power(2, i))
            self.tile_to_texture_pixelratio[i] = tile_to_texture_pixelratioX

    def request_url(self, zoom: int, x: int, y: int) -> str:
        """
        Return the url to request the tile as a string
        """
        return self.provider.build_url(z=zoom, x=x, y=y)

    def tile_shape(self, zoom: int) -> tuple[int, int]:
        """
        Return the shape (width, height) in pixels of a tile at the given zoom level
        """
        return (256, 256)

    def tilematrix_xy_to_tilematrix_crs(self, zoom: int, x: int, y: int) -> tuple[float, float, float, float]:
        """
        Return the rect of the tile in self.crs coordinates as left, bottom, right, top
        """
        if zoom < 0 or zoom > self.provider.max_zoom:
            raise ValueError(f'zoom {zoom} is < 0 or > {self.provider.max_zoom} max provider zoom')
        if x < 0 or x >= int(np.power(2, zoom)):
            raise ValueError(f'x {x} is < 0 or >= 2^zoom = {int(np.power(2, zoom))}')
        if y < 0 or y >= int(np.power(2, zoom)):
            raise ValueError(f'y {y} is < 0 or >= 2^zoom = {int(np.power(2, zoom))}')
        left, right = x_to_long([x, x+1], zoom)
        top, bottom = y_to_lat([y, y+1], zoom)
        return left, bottom, right, top

    def main_crs_to_tilematrix_xy(self, left: float, bottom: float, right: float, top: float,
                                  zoom: int) -> tuple[int, int, int, int]:
        """
        Return the rect of tiles in xy coordinates as left, bottom, right, top
        """
        if zoom < 0 or zoom > self.provider.max_zoom:
            raise ValueError(f'zoom {zoom} is < 0 or > {self.provider.max_zoom} max provider zoom')
        # transform into coordinates in the tilematrix crs (long lat)
        left, bottom, right, top = rasterio.warp.transform_bounds(self.main_crs, self.crs,
                                                                  left, bottom, right, top)
        left, right = long_to_x([left, right], zoom)
        top, bottom = lat_to_y([top, bottom], zoom)
        return left, bottom, right, top

    def to_dict(self, project_path: pathlib.Path) -> dict[str, Any]:
        output: dict[str, Any] = super().to_dict(project_path)
        output["provider"] = dict(self.provider)
        return output

    @classmethod
    def from_dict(cls, input_dict: dict[str, Any], map_model: "MapModel", map_widget: "AbstractMapView") -> "XYZLayer":
        assert input_dict["kind"] == cls.kind
        name = input_dict.get("name", "XYZ layer")
        try:
            provider = xyz.providers.flatten()[input_dict["provider"]["name"]]
        except:
            provider = xyz.TileProvider(input_dict["provider"])
        layer = XYZLayer(name, map_model, map_widget, provider)
        if "visible" in input_dict:
            layer.visible = input_dict["visible"]
        if "alpha" in input_dict:
            layer.alpha = input_dict["alpha"]
        return layer


def long_to_x(long: Union[float, list[float]], zoom: int) -> Union[int, list[int]]:
    # see https://wiki.openstreetmap.org/wiki/Slippy_map_tilenames#Mathematics
    if isinstance(long, (int, float)):
        result = int(np.floor(np.power(2, zoom) * (long + 180.0) / 360.0))
        return min(np.power(2, zoom) - 1, max(0, result))
    return [long_to_x(_, zoom) for _ in long]


def x_to_long(x: Union[int, list[int]], zoom: int) -> Union[float, list[float]]:
    # see https://wiki.openstreetmap.org/wiki/Slippy_map_tilenames#Mathematics
    if isinstance(x, (int)):
        return x / np.power(2, zoom) * 360.0 - 180.0
    return [x_to_long(_, zoom) for _ in x]


def lat_to_y(lat: Union[float, list[float]], zoom: int) -> Union[int, list[int]]:
    # see https://wiki.openstreetmap.org/wiki/Slippy_map_tilenames#Mathematics
    if isinstance(lat, (int, float)):
        result = int(np.floor(np.power(2, zoom) *
                              (1.0 - np.asinh(np.tan(np.radians(lat))) / np.pi) / 2.0))
        return min(np.power(2, zoom) - 1, max(0, result))
    return [lat_to_y(_, zoom) for _ in lat]


def y_to_lat(y: Union[int, list[int]], zoom: int) -> Union[float, list[float]]:
    # see https://wiki.openstreetmap.org/wiki/Slippy_map_tilenames#Mathematics
    if isinstance(y, (int)):
        return np.degrees(np.atan(np.sinh(np.pi * (1 - 2 * y / np.power(2, zoom)))))
    return [y_to_lat(_, zoom) for _ in y]


class OpenStreetMapLayer(XYZLayer):

    kind = "OpenStreetMap Layer"

    def __init__(self, model: "MapModel", map_widget: "AbstractMapView", name: str = "OpenStreetMap"):
        super().__init__(name, model, map_widget, xyz.providers.OpenStreetMap.Mapnik)

    to_dict = OpenGLLayer.to_dict

    @classmethod
    def from_dict(cls, input_dict: dict[str, Any], map_model: "MapModel", map_widget: "AbstractMapView") -> "OpenStreetMapLayer":
        assert input_dict["kind"] == cls.kind
        name = input_dict.get("name", "OpenStreetMap")
        layer = OpenStreetMapLayer(map_model, map_widget, name=name)
        if "visible" in input_dict:
            layer.visible = input_dict["visible"]
        if "alpha" in input_dict:
            layer.alpha = input_dict["alpha"]
        return layer


class XYZLayerDialog(QDialog):

    def __init__(self, parent: Optional[QWidget] = None):
        super().__init__(parent)
        self.provider: Optional[xyz.lib.TileProvider] = None
        self.setWindowModality(Qt.WindowModality.ApplicationModal)
        self.setModal(True)
        self.setWindowTitle("Create XYZ Layer")
        # labels
        self.provider_label = QLabel("Provider:")
        self.url_label = QLabel("URL:")
        self.min_zoom_label = QLabel("Min Zoom level:")
        self.max_zoom_label = QLabel("Max Zoom level:")
        # inputs
        self.provider_input = QLineEdit("Custom")
        completer_list = list(xyz.providers.flatten().keys())
        completer_list.append("Custom")
        self.provider_completer = QCompleter(completer_list)
        self.provider_completer.setCaseSensitivity(Qt.CaseSensitivity.CaseInsensitive)
        self.provider_input.setCompleter(self.provider_completer)
        self.provider_input.textChanged.connect(self.set_provider)
        self.url_input = QLineEdit()
        self.url_input.setPlaceholderText(r"http://example.com/{z}/{x}/{y}.png")
        self.url_input.textChanged.connect(self.set_custom)
        self.min_zoom_input = QSpinBox()
        self.min_zoom_input.setMinimum(0)
        self.min_zoom_input.valueChanged.connect(self.set_custom)
        self.max_zoom_input = QSpinBox()
        self.max_zoom_input.setMinimum(0)
        self.max_zoom_input.valueChanged.connect(self.set_custom)
        self.api_key_groupbox = QGroupBox("Authentication")
        self.api_key_label = QLabel("Api Key:")
        self.api_key_input = QLineEdit()
        self.api_key_input.textChanged.connect(self.test_enable_ok)
        self.api_key_layout = QHBoxLayout()
        self.api_key_layout.addWidget(self.api_key_label)
        self.api_key_layout.addWidget(self.api_key_input)
        self.api_key_groupbox.setLayout(self.api_key_layout)
        self.api_key_groupbox.setCheckable(True)
        self.api_key_groupbox.setChecked(False)
        self.api_key_groupbox.clicked.connect(self.set_custom)
        # buttons
        self.button_box = QDialogButtonBox()
        self.cancel_button = self.button_box.addButton(QDialogButtonBox.StandardButton.Cancel)
        self.cancel_button.clicked.connect(self.reject)
        self.ok_button = self.button_box.addButton(QDialogButtonBox.StandardButton.Ok)
        self.ok_button.clicked.connect(self.accept)
        # main layout
        self.main_layout = QGridLayout()
        self.main_layout.addWidget(self.provider_label, 0, 0)
        self.main_layout.addWidget(self.provider_input, 0, 1)
        self.main_layout.addWidget(self.url_label, 1, 0)
        self.main_layout.addWidget(self.url_input, 1, 1)
        self.main_layout.addWidget(self.min_zoom_label, 2, 0)
        self.main_layout.addWidget(self.min_zoom_input, 2, 1)
        self.main_layout.addWidget(self.max_zoom_label, 3, 0)
        self.main_layout.addWidget(self.max_zoom_input, 3, 1)
        self.main_layout.addWidget(self.api_key_groupbox, 4, 0, 1, 2)
        self.main_layout.addWidget(self.button_box, 5, 0, 1, 2)
        self.setLayout(self.main_layout)
        self.test_enable_ok()

    @Slot(str)
    def set_provider(self, provider_name: str) -> None:
        providers = xyz.providers.flatten()
        if provider_name in providers.keys():
            self.provider = providers[provider_name]
            # QSignalBlocker to prevent valueChanged signals from being fired
            with QSignalBlocker(self.url_input):
                self.url_input.setText(self.provider["url"])
            with QSignalBlocker(self.min_zoom_input):
                self.min_zoom_input.setValue(self.provider.get("min_zoom", 0))
            with QSignalBlocker(self.max_zoom_input):
                self.max_zoom_input.setValue(self.provider.get("max_zoom", 18))
            self.api_key_input.setText("")
            with QSignalBlocker(self.api_key_groupbox):
                if self.provider.requires_token():
                    self.api_key_groupbox.setChecked(True)
                else:
                    self.api_key_groupbox.setChecked(False)
            self.test_enable_ok()
        elif self.provider is not None:
            self.set_custom()

    @Slot()
    def set_custom(self) -> None:
        self.provider = None
        self.provider_input.setText("Custom")
        self.test_enable_ok()

    @Slot()
    def test_enable_ok(self) -> None:
        enable_ok = True
        if self.min_zoom_input.value() > self.max_zoom_input.value():
            enable_ok = False
        if self.provider is not None:
            if self.provider.requires_token():
                if self.api_key_input.text() == "" or not self.api_key_groupbox.isChecked():
                    enable_ok = False
        else:
            if self.url_input.text() == "":
                enable_ok = False
            if self.api_key_groupbox.isChecked() and self.api_key_input.text() == "":
                enable_ok = False
        self.ok_button.setEnabled(enable_ok)

    def accept(self) -> None:
        if self.provider is not None:
            if self.api_key_groupbox.isChecked():
                self.provider["apiKey"] = self.api_key_input.text()
        else:
            self.provider = xyz.TileProvider(name=self.provider_input.text(),
                                             url=self.url_input.text(),
                                             attribution=self.provider_input.text(),
                                             min_zoom=self.min_zoom_input.value(),
                                             max_zoom=self.max_zoom_input.value())
        super().accept()

    def sizeHint(self) -> QSize:
        return QSize(768, super().sizeHint().height())

    def keyPressEvent(self, e) -> None:
        # ignore "Enter" key press that would otherwise trigger ok button
        if e.key() in (Qt.Key.Key_Return, Qt.Key.Key_Enter):
            e.accept()
        else:
            super().keyPressEvent(e)



#!/usr/bin/env python3
""" PlotModel

This module contains the PlotModel class pertaining to the management of data
selected by the user to be displayed in plot views.

Works with the PlotView module, as in a Model/View architecture.

"""

from typing import Any, Optional, Union, Iterable, overload

import numpy as np

from PySide6.QtCore import (
    Qt, QObject, Slot, Signal, QAbstractProxyModel, QModelIndex, QPersistentModelIndex,
    QAbstractItemModel
)

from insarviz.Loader import Loader

from insarviz.map.layers.LayerModel import LayerModel

from insarviz.map.layers.SelectionLayer import SelectionItem

from insarviz.Roles import Roles


# PlotModel class ##################################################################################

class PlotModel(QObject):

    closed = Signal()
    updated_pointer_info = Signal(tuple, np.ndarray)
    points_added = Signal(int, int)
    points_removed = Signal(int, int)
    points_updated = Signal(int, int, list)
    profiles_added = Signal(int, int)
    profiles_removed = Signal(int, int)
    profiles_updated = Signal(int, int, list)
    references_added = Signal(int, int)
    references_removed = Signal(int, int)
    references_updated = Signal(int, int, list)

    def __init__(self, loader: Loader):
        """
        Model for plots
        get and transform date/band values (timestamps, datetime or int)
        get and prepare data for plots

        Note that band index in data structures start at 0 like in usual python, whereas band index
        start at 1 in MapModel and in gdal, MapModel/gdal indexes are transformed using:
        self.loader.dataset.indexes.index(band_index)

        Parameters
        ----------
        loader : Loader

        Returns
        -------
        None.
        """
        super().__init__()
        self.loader: Loader = loader
        self.loader.data_invalidated.connect(self.on_data_invalidated)
        # data for graphs:
        self.points: ProxySelectionModel = ProxySelectionModel()
        #  list of arrays: points_data[point][date]
        self.points_data: list[np.ndarray] = []
        self.profiles: ProxySelectionModel = ProxySelectionModel()
        # list of arrays profiles_data[profile][point along profile, date]
        self.profiles_data: list[np.ndarray] = []
        self.references: ProxySelectionModel = ProxySelectionModel()
        # list of arrays references_data[reference][date]
        self.references_data: list[np.ndarray] = []
        # signals and slots
        self.points.rowsInserted.connect(lambda _, first, last: self.add_points(first, last))
        self.points.rowsRemoved.connect(lambda _, first, last: self.remove_points(first, last))
        self.points.dataChanged.connect(lambda topLeft, bottomRight, roles:
                                        self.update_points(topLeft.row(), bottomRight.row(), roles))
        self.profiles.rowsInserted.connect(lambda _, first, last: self.add_profiles(first, last))
        self.profiles.rowsRemoved.connect(lambda _, first, last: self.remove_profiles(first, last))
        self.profiles.dataChanged.connect(lambda topLeft, bottomRight, roles: self.update_profiles(
            topLeft.row(), bottomRight.row(), roles))
        self.references.rowsInserted.connect(
            lambda _, first, last: self.add_references(first, last))
        self.references.rowsRemoved.connect(
            lambda _, first, last: self.remove_references(first, last))
        self.references.dataChanged.connect(lambda topLeft, bottomRight, roles:
                                            self.update_references(topLeft.row(), bottomRight.row(),
                                                                   roles))

    def close(self) -> None:
        self.closed.emit()
        self.points.set_root_index()
        self.points_data = []
        self.profiles.set_root_index()
        self.profiles_data = []
        self.references.set_root_index()
        self.references_data = []

    @Slot()
    def on_data_invalidated(self) -> None:
        self.update_points(0, len(self.points_data)-1, [Roles.ComputeDataRole])
        self.update_profiles(0, len(self.profiles_data)-1, [Roles.ComputeDataRole])
        self.update_references(0, len(self.references_data)-1, [Roles.ComputeDataRole])

    # connected to LayerModel.selection_initialized
    @Slot(object)
    def on_selection_init(self, layer_model: LayerModel) -> None:
        assert layer_model.selection is not None
        selection_index = layer_model.index(
            layer_model.selection.child_number(), 0, QModelIndex())
        points_folder_index = layer_model.index(
            layer_model.selection.points_folder.child_number(), 0, selection_index)
        profiles_folder_index = layer_model.index(
            layer_model.selection.profiles_folder.child_number(), 0, selection_index)
        references_folder_index = layer_model.index(
            layer_model.selection.references_folder.child_number(), 0, selection_index)
        self.points.setSourceModel(layer_model)
        self.points.set_root_index(points_folder_index)
        self.profiles.setSourceModel(layer_model)
        self.profiles.set_root_index(profiles_folder_index)
        self.references.setSourceModel(layer_model)
        self.references.set_root_index(references_folder_index)
        if layer_model.rowCount(points_folder_index):
            self.add_points(0, layer_model.rowCount(points_folder_index)-1)
        if layer_model.rowCount(profiles_folder_index):
            self.add_profiles(0, layer_model.rowCount(profiles_folder_index)-1)
        if layer_model.rowCount(references_folder_index):
            self.add_references(0, layer_model.rowCount(references_folder_index)-1)

    @Slot(int, int)
    def add_points(self, first: int, last: int) -> None:
        """
        Slot connected to self.points.rowsInserted, add to self.points_data the data corresponding
        to the rows inserted in self.points. Emits self.points_added signal.

        Parameters
        ----------
        first : int
            index of first point/row added
        last : int
            index of last point/row added
        """
        indexes = [self.points.index(i, 0) for i in range(first, last+1)]
        points = [self.points.get_selection_item(i) for i in indexes]
        self.points_data[first:first] = [self.loader.load_profile_window(*i.get_rect()[0])
                                         for i in points]
        self.points_added.emit(first, last)

    @Slot(int, int)
    def remove_points(self, first: int, last: int) -> None:
        """
        Slot connected to self.points.rowsRemoved, remove from self.points_data the data
        corresponding to the rows removed from self.points. Emits self.points_removed signal.

        Parameters
        ----------
        first : int
            index of first point/row removed
        last : int
            index of last point/row removed
        """
        del self.points_data[first:last+1]
        self.points_removed.emit(first, last)

    @Slot(int, int, list)
    def update_points(self, first: int, last: int, roles: list[int]) -> None:
        """
        Slot connected to self.points.dataChanged, update in self.points_data the data corresponding
        to the points that have been changed in self.points. Emits self.points_updated signal.

        Parameters
        ----------
        first : int
            index of first point/row whose data have been modified
        last : int
            index of last point/row whose data have been modified
        roles : list of int
            list of the ItemDataRoles that have been modified
            see https://doc.qt.io/qt-5/qt.html#ItemDataRole-enum and insarviz.Roles
        """
        if Roles.ComputeDataRole in roles:
            indexes = [self.points.index(i, 0) for i in range(first, last+1)]
            points = [self.points.get_selection_item(i) for i in indexes]
            self.points_data[first:last+1] = [self.loader.load_profile_window(*i.get_rect()[0])
                                              for i in points]
        self.points_updated.emit(first, last, roles)

    @Slot(int, int)
    def add_profiles(self, first: int, last: int) -> None:
        """
        Slot connected to self.profiles.rowsInserted, add to self.profiles_data the data
        corresponding to the rows inserted in self.profiles. Emits self.profiles_added signal.

        Parameters
        ----------
        first : int
            index of first profile/row added
        last : int
            index of last profile/row added
        """
        indexes = [self.profiles.index(i, 0) for i in range(first, last+1)]
        profiles = [self.profiles.get_selection_item(i) for i in indexes]
        self.profiles_data[first:first] = [np.array([self.loader.load_profile_window(*rect)
                                                    for rect in i.get_rect()]) for i in profiles]
        self.profiles_added.emit(first, last)

    @Slot(int, int)
    def remove_profiles(self, first: int, last: int) -> None:
        """
        Slot connected to self.profiles.rowsRemoved, remove from self.profiles_data the data
        corresponding to the rows removed from self.profiles. Emits self.profiles_removed signal.

        Parameters
        ----------
        first : int
            index of first profile/row removed
        last : int
            index of last profile/row removed
        """
        del self.profiles_data[first:last+1]
        self.profiles_removed.emit(first, last)

    @Slot(int, int, list)
    def update_profiles(self, first: int, last: int, roles: list[int]) -> None:
        """
        Slot connected to self.profiles.dataChanged, update in self.profiles_data the data
        corresponding to the profiles that have been changed in self.profiles. Emits
        self.profiles_updated signal.

        Parameters
        ----------
        first : int
            index of first profile/row whose data have been modified
        last : int
            index of last profile/row whose data have been modified
        roles : list of int
            list of the ItemDataRoles that have been modified
            see https://doc.qt.io/qt-5/qt.html#ItemDataRole-enum and insarviz.Roles
        """
        if Roles.ComputeDataRole in roles:
            indexes = [self.profiles.index(i, 0) for i in range(first, last+1)]
            profiles = [self.profiles.get_selection_item(i) for i in indexes]
            self.profiles_data[first:last+1] = [np.array([self.loader.load_profile_window(*rect)
                                                for rect in i.get_rect()]) for i in profiles]
        self.profiles_updated.emit(first, last, roles)

    @Slot(int, int)
    def add_references(self, first: int, last: int) -> None:
        """
        Slot connected to self.references.rowsInserted, add to self.references_data the data
        corresponding to the rows inserted in self.references. Emits self.references_added signal.

        Parameters
        ----------
        first : int
            index of first reference/row added
        last : int
            index of last reference/row added
        """
        indexes = [self.references.index(i, 0) for i in range(first, last+1)]
        references = [self.references.get_selection_item(i) for i in indexes]
        self.references_data[first:first] = [self.loader.load_profile_window(*i.get_rect()[0])
                                             for i in references]
        self.references_added.emit(first, last)

    @Slot(int, int)
    def remove_references(self, first: int, last: int) -> None:
        """
        Slot connected to self.references.rowsRemoved, remove from self.references_data the data
        corresponding to the rows removed from self.references. Emits self.references_removed
        signal.

        Parameters
        ----------
        first : int
            index of first reference/row removed
        last : int
            index of last reference/row removed
        """
        del self.references_data[first:last+1]
        self.references_removed.emit(first, last)

    @Slot(int, int, list)
    def update_references(self, first: int, last: int, roles: list[int]) -> None:
        """
        Slot connected to self.references.dataChanged, update in self.references_data the data
        corresponding to the points that have been changed in self.references. Emits
        self.references_updated signal.

        Parameters
        ----------
        first : int
            index of first reference/row whose data have been modified
        last : int
            index of last reference/row whose data have been modified
        roles : list of int
            list of the ItemDataRoles that have been modified
            see https://doc.qt.io/qt-5/qt.html#ItemDataRole-enum and insarviz.Roles
        """
        if Roles.ComputeDataRole in roles:
            indexes = [self.references.index(i, 0) for i in range(first, last+1)]
            references = [self.references.get_selection_item(i) for i in indexes]
            self.references_data[first:first] = [self.loader.load_profile_window(*i.get_rect()[0])
                                                 for i in references]
        self.references_updated.emit(first, last, roles)

    @Slot(tuple, int)
    def update_pointer(self, pos: tuple, date_index: Optional[int]) -> None:
        """
        load current pointer data (for plots)
        launched by mouseMoveEvent on MapView
        date_index correspond to MapModel.current_band_index (starting from 1 like in gdal)
        """
        if pos == ():
            self.updated_pointer_info.emit((), np.array([]))
        else:
            assert self.loader.dataset is not None
            assert date_index is not None and date_index in self.loader.dataset.indexes
            # then pos = (i,j)
            (i, j) = pos
            # load data for all dates at current pointer's position
            pointer_data = self.loader.load_profile_window(i, j)
            # emit (x, y, date, value)
            # need to take date_index-1 because date_index starts from 1 like in gdal
            self.updated_pointer_info.emit(
                (i, j, self.loader.dates[date_index-1], pointer_data[date_index-1]), pointer_data)


# ProxySelectionModel class ########################################################################

class ProxySelectionModel(QAbstractProxyModel):

    # transform a TreeModel into a list of root_index's children
    # (children at depth >1 are filtered out)
    # return the list of the selection items in a folder when used with root_index=SelectionFolder
    # also replace Qt.CheckStateRole by SelectionItem.ShowCurveRole

    def __init__(self, source_model: Optional[QAbstractItemModel] = None,
                 root_index: Optional[QModelIndex] = None):
        super().__init__()
        if source_model is not None:
            self.setSourceModel(source_model)
        self.root_index: Optional[QPersistentModelIndex]
        self.set_root_index(root_index)

    # methods reimplemented from QAbstractProxyModel

    def setSourceModel(self, source_model: QAbstractItemModel) -> None:
        # pylint: disable=missing-function-docstring, invalid-name
        if self.sourceModel() is not None:
            self.sourceModel().rowsAboutToBeInserted.disconnect(self.source_rows_about_inserted)
            self.sourceModel().rowsInserted.disconnect(self.source_rows_inserted)
            self.sourceModel().rowsAboutToBeRemoved.disconnect(self.source_rows_about_removed)
            self.sourceModel().rowsRemoved.disconnect(self.source_rows_removed)
            self.sourceModel().dataChanged.disconnect(self.source_data_changed)
            self.sourceModel().modelReset.disconnect(self.set_root_index)
        super().setSourceModel(source_model)
        self.sourceModel().rowsAboutToBeInserted.connect(self.source_rows_about_inserted)
        self.sourceModel().rowsInserted.connect(self.source_rows_inserted)
        self.sourceModel().rowsAboutToBeRemoved.connect(self.source_rows_about_removed)
        self.sourceModel().rowsRemoved.connect(self.source_rows_removed)
        self.sourceModel().dataChanged.connect(self.source_data_changed)
        self.sourceModel().modelReset.connect(self.set_root_index)

    def rowCount(self, parent: QModelIndex = QModelIndex()) -> int:
        # pylint: disable=missing-function-docstring, invalid-name
        if parent == QModelIndex():
            if self.root_index is None or not self.root_index.isValid():
                return 0
            return self.sourceModel().rowCount(QModelIndex(self.root_index))
        return 0

    def columnCount(self, parent: QModelIndex = QModelIndex()) -> int:
        # pylint: disable=missing-function-docstring, invalid-name, unused-argument
        return 1

    def index(self, row: int, column: int, parent: QModelIndex = QModelIndex()) -> QModelIndex:
        # pylint: disable=missing-function-docstring
        if parent.isValid() or self.rowCount(parent) <= row or self.columnCount(parent) <= column:
            return QModelIndex()
        assert self.root_index is not None and self.root_index.isValid()
        pointer = self.sourceModel().index(row, column,
                                           QModelIndex(self.root_index)).internalPointer()
        return self.createIndex(row, column, pointer)

    # see https://mypy.readthedocs.io/en/stable/more_types.html#function-overloading
    @overload
    def parent(self, child: QModelIndex) -> QModelIndex: ...

    @overload
    def parent(self) -> QObject: ...

    def parent(self, child: Optional[QModelIndex] = None) -> Union[QObject, QModelIndex]:
        # pylint: disable=missing-function-docstring
        if child is None:
            return super().parent()
        return QModelIndex()

    def data(self, index: QModelIndex, role: int = Qt.ItemDataRole.DisplayRole) -> Any:
        # pylint: disable=missing-function-docstring
        if role == Qt.ItemDataRole.CheckStateRole:
            # use SelectionItem.show_curve field
            return super().data(index, SelectionItem.ShowCurveRole)
        return super().data(index, role)

    def setData(self, index: QModelIndex, value: Any, role: int = Qt.ItemDataRole.EditRole) -> bool:
        # pylint: disable=missing-function-docstring, invalid-name
        if role == Qt.ItemDataRole.CheckStateRole:
            # use SelectionItem.show_curve field
            return super().setData(index, value, SelectionItem.ShowCurveRole)
        if role in (Roles.ProfileTemporalRole, Roles.ProfileSpatialRole):
            return super().setData(index, value, role)
        return False

    def flags(self, index: QModelIndex) -> Qt.ItemFlags:
        # pylint: disable=missing-function-docstring
        return super().flags(index) & (~Qt.ItemFlag.ItemIsEditable)  # remove editable flag

    def mapFromSource(self, sourceIndex: QModelIndex) -> QModelIndex:
        # pylint: disable=missing-function-docstring, invalid-name
        if not sourceIndex.isValid():
            return QModelIndex()
        if sourceIndex == self.root_index:
            return QModelIndex()
        if sourceIndex.parent() == self.root_index:
            return self.index(sourceIndex.row(), sourceIndex.column(), QModelIndex())
        return QModelIndex()

    def mapToSource(self, proxyIndex: QModelIndex) -> QModelIndex:
        # pylint: disable=missing-function-docstring, invalid-name
        if self.root_index is None or not self.root_index.isValid():
            return QModelIndex()
        if not proxyIndex.isValid():
            return QModelIndex(self.root_index)
        return self.sourceModel().index(proxyIndex.row(), proxyIndex.column(),
                                        QModelIndex(self.root_index))

    # custom methods
    @Slot()
    def set_root_index(self, index: Optional[QModelIndex] = None) -> None:
        if index is None:
            self.root_index = None
        else:
            self.root_index = QPersistentModelIndex(index)

    def get_selection_item(self, index: QModelIndex) -> SelectionItem:
        layer_model = self.sourceModel()
        assert isinstance(layer_model, LayerModel)
        item = layer_model.get_item(index)
        assert isinstance(item, SelectionItem)
        return item

    @Slot(QModelIndex, int, int)
    def source_rows_about_inserted(self, parent: QModelIndex, first: int, last: int) -> None:
        # pylint: disable=missing-function-docstring
        if parent == self.root_index:
            self.beginInsertRows(self.mapFromSource(parent), first, last)

    @Slot(QModelIndex, int, int)
    def source_rows_inserted(self, parent: QModelIndex, first: int, last: int) -> None:
        # pylint: disable=missing-function-docstring, unused-argument
        if parent == self.root_index:
            self.endInsertRows()

    @Slot(QModelIndex, int, int)
    def source_rows_about_removed(self, parent: QModelIndex, first: int, last: int) -> None:
        # pylint: disable=missing-function-docstring
        if parent == self.root_index:
            self.beginRemoveRows(self.mapFromSource(parent), first, last)

    @Slot(QModelIndex, int, int)
    def source_rows_removed(self, parent: QModelIndex, first: int, last: int) -> None:
        # pylint: disable=unused-argument, missing-function-docstring
        if parent == self.root_index:
            self.endRemoveRows()

    #  TODO cannot find correct Slot signature:
    # @Slot(QModelIndex, QModelIndex, list) not working
    def source_data_changed(self, top_left: QModelIndex, bottom_right: QModelIndex,
                            roles: Iterable[int]) -> None:
        # pylint: disable=missing-function-docstring
        if top_left.parent() == self.root_index and bottom_right.parent() == self.root_index:
            if top_left.column() == bottom_right.column() == 0:
                self.dataChanged.emit(self.mapFromSource(top_left),
                                      self.mapFromSource(bottom_right), roles)

    @Slot()
    def show_all_curves(self) -> None:
        for i in range(self.rowCount(QModelIndex())):
            index = self.index(i, 0, QModelIndex())
            self.setData(index, Qt.CheckState.Checked, Qt.ItemDataRole.CheckStateRole)

    @Slot()
    def hide_all_curves(self) -> None:
        for i in range(self.rowCount(QModelIndex())):
            index = self.index(i, 0, QModelIndex())
            self.setData(index, Qt.CheckState.Unchecked, Qt.ItemDataRole.CheckStateRole)
